Package: MultiSpline
Type: Package
Title: Spline-Based Nonlinear Modeling for Multilevel and Longitudinal
        Data
Version: 0.2.0
Authors@R: person(given = "Subir",
                  family = "Hait",
                  email = "haitsubi@msu.edu",
                  role = c("aut", "cre"),
                  comment = c(ORCID = "0009-0004-9871-9677"))
Description: Provides a unified framework for fitting, predicting, and
    interpreting nonlinear relationships in single-level, multilevel, and
    longitudinal regression models. Flexible functional forms are supported
    using natural cubic splines ('splines'), B-splines ('splines'), and GAM
    smooths ('mgcv'). Supports two-way and nested clustering via 'lme4',
    automatic knot selection by AIC or BIC, multilevel R-squared
    decomposition (Nakagawa-Schielzeth marginal and conditional R-squared
    with level-specific variance partitioning), a postestimation suite
    returning first and second derivatives with confidence bands, turning
    points and inflection regions, and a model comparison workflow
    contrasting linear, polynomial, and spline fits by AIC, BIC, and
    likelihood-ratio tests. Cluster heterogeneity in nonlinear effects is
    supported via random-slope spline terms.
Depends: R (>= 4.2.0)
Imports: stats, lme4, mgcv, dplyr, ggplot2, rlang, splines
Suggests: lmerTest, knitr, rmarkdown, reformulas, numDeriv
License: GPL-3
Encoding: UTF-8
Language: en-US
URL: https://github.com/causalfragility-lab/MultiSpline
BugReports: https://github.com/causalfragility-lab/MultiSpline/issues
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-16 03:52:32 UTC; Subir
Author: Subir Hait [aut, cre] (ORCID: <https://orcid.org/0009-0004-9871-9677>)
Maintainer: Subir Hait <haitsubi@msu.edu>
Repository: CRAN
Date/Publication: 2026-04-16 07:02:39 UTC
Built: R 4.6.0; ; 2026-04-26 14:10:22 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,nl_compare)
S3method(print,nl_fit)
S3method(print,nl_r2)
S3method(print,summary_nl_fit)
S3method(summary,nl_fit)
export(nl_compare)
export(nl_derivatives)
export(nl_fit)
export(nl_het)
export(nl_icc)
export(nl_knots)
export(nl_plot)
export(nl_predict)
export(nl_r2)
export(nl_summary)
export(nl_turning_points)
importFrom(rlang,"%||%")
importFrom(rlang,.data)
importFrom(splines,bs)
importFrom(splines,ns)

# MultiSpline v2.0.0

**Spline-Based Nonlinear Modeling for Multilevel and Longitudinal Data**

> Author: Subir Hait · Michigan State University · haitsubi@msu.edu

---

## Overview

MultiSpline provides a unified interface for fitting, interpreting, and
visualising nonlinear relationships in single-level and multilevel regression
models using natural cubic splines, B-splines, or GAM smooths.

Version 2 is a major upgrade. Every item raised in the Moran correspondence
has been addressed, and the package now offers a complete **estimation →
interpretation → diagnostics** framework:

| Layer | Functions |
|-------|-----------|
| **Estimation** | `nl_fit()` · `nl_knots()` |
| **Prediction** | `nl_predict()` |
| **Interpretation** | `nl_derivatives()` · `nl_turning_points()` · `nl_plot()` |
| **Model selection** | `nl_compare()` |
| **Variance explained** | `nl_r2()` · `nl_icc()` |
| **Cluster heterogeneity** | `nl_het()` |

---

## What's New in v2

### 1. Two-way and Nested Clustering  *(Moran request)*

```r
# Cross-classified (e.g. students nested in both schools and districts)
fit <- nl_fit(data = df, y = "score", x = "SES",
              cluster = c("student_id", "school_id"),
              nested  = FALSE)

# Nested (students within schools)
fit <- nl_fit(data = df, y = "score", x = "SES",
              cluster = c("student_id", "school_id"),
              nested  = TRUE)
```

Internally this generates `(1 | student_id) + (1 | school_id)` for
cross-classified models and `(1 | student_id/school_id)` for nested models,
passed to `lme4::lmer()` / `lme4::glmer()`.

---

### 2. Confidence Intervals for Spline Curves  *(Moran request)*

All model types now return proper CI bands:

| Model | CI method |
|-------|-----------|
| `lm` | Analytical (standard `predict.lm`) |
| `gam` | Analytical (`mgcv`) |
| `lmerMod` | Delta method via fixed-effects variance–covariance matrix |
| `glmerMod` | Delta method on link scale + back-transformation (default) or parametric bootstrap (`glmer_ci = "boot"`) |

```r
pred <- nl_predict(fit, se = TRUE, level = 0.95)
nl_plot(pred_df = pred, x = "age", show_ci = TRUE)
```

---

### 3. Automatic Knot / df Selection

```r
# Explore and select df explicitly
ks <- nl_knots(data = df, y = "score", x = "age",
               df_range = 2:10, criterion = "AIC")
ks$best_df   # → e.g. 5

# Or let nl_fit do it automatically
fit <- nl_fit(data = df, y = "score", x = "age",
              df = "auto", df_criterion = "AIC")
```

A diagnostic plot of AIC/BIC vs df is produced automatically.

---

### 4. Multilevel R² Decomposition  *(Moran request — r2_mlm)*

```r
nl_r2(fit)
```

**Output (LMM example):**
```
=== MultiSpline R² Decomposition ===
Model type: LMM

  Marginal  R²m = 0.1823   (fixed effects only)
  Conditional R²c = 0.4761  (fixed + all random effects)

Variance Partition (r2_mlm-style):
     Component  Variance  Proportion
Fixed effects    1.4221      0.1823
student_id       1.8340      0.2350
school_id        0.8762      0.1124
Residual         3.6590      0.4690
```

Follows the Nakagawa-Schielzeth (2013) and Nakagawa-Johnson-Schielzeth (2017)
formulas for marginal and conditional R². The level-specific variance
partition mirrors the `r2_mlm` / Rights-Sterba (2019) approach.

---

### 5. Derivative-Based Interpretation Framework

```r
pred  <- nl_predict(fit, x_seq = seq(18, 80, length.out = 200))
deriv <- nl_derivatives(pred, x = "age")

# Visualise slope (first derivative) with CI band
nl_plot(deriv_df = deriv, x = "age", type = "slope")

# Visualise curvature (second derivative)
nl_plot(deriv_df = deriv, x = "age", type = "curvature")

# Trajectory + slope side by side
nl_plot(pred_df = pred, deriv_df = deriv, x = "age", type = "combo")
```

**Turning points and inflection regions:**

```r
tp <- nl_turning_points(deriv, x = "age")
tp$turning_points      # local maxima and minima with x location and type
tp$inflection_regions  # where concavity changes
tp$slope_regions       # contiguous increasing / decreasing stretches

# Overlay turning points on trajectory plot
nl_plot(pred_df = pred, x = "age",
        show_turning_points = TRUE,
        turning_points      = tp$turning_points)
```

---

### 6. Model Comparison Workflow

```r
nl_compare(fit)                           # default: linear + poly(2) + poly(3) + spline
nl_compare(fit, polynomial_degrees = 2:5) # custom comparators
```

**Output:**
```
=== MultiSpline Model Comparison ===

    Model    AIC    BIC  LogLik  npar  Deviance  LRT_vs_linear  LRT_p
   Linear 1842.1 1863.4  -916.1     5    1020.4             NA     NA
  Poly(2) 1830.5 1855.6  -908.2     6     992.3          15.74  <0.001
  Poly(3) 1825.3 1854.3  -903.6     7     978.1          25.00  <0.001
   Spline 1818.7 1851.5  -898.3     9     961.2          35.68  <0.001

  Best model by AIC: Spline
```

---

### 7. Cluster Heterogeneity in Nonlinear Effects

```r
nl_het(fit, n_clusters_plot = 50)
```

Plots cluster-specific predicted trajectories (BLUPs) against the
population-mean curve, and performs a likelihood-ratio test comparing
random-slope vs random-intercept models to test whether the nonlinear
effect genuinely varies across clusters.

---

### 8. B-spline Basis

```r
fit_bs <- nl_fit(data = df, y = "score", x = "age",
                 method = "bs", df = 6, bs_degree = 3)
```

---

### 9. Random Spline Slopes

Allows the nonlinear trajectory to vary across clusters:

```r
fit_rs <- nl_fit(data = df, y = "score", x = "age",
                 cluster = "id", random_slope = TRUE)
```

---

## Installation

```r
# From source
install.packages("path/to/MultiSpline_0.2.0.tar.gz", repos = NULL, type = "source")

# Dependencies (install first if needed)
install.packages(c("lme4", "mgcv", "dplyr", "ggplot2", "rlang", "splines"))
# Optional (for p-values in LMM)
install.packages("lmerTest")
```

---

## Quick-Start Example

```r
library(MultiSpline)

# 1. Select df automatically
ks <- nl_knots(data = nlme::Orthodont, y = "distance",
               x = "age", df_range = 2:8)

# 2. Fit multilevel model with two-way clustering and auto df
fit <- nl_fit(
  data     = nlme::Orthodont,
  y        = "distance",
  x        = "age",
  cluster  = "Subject",
  df       = "auto",
  controls = NULL
)
fit   # prints postestimation menu

# 3. Coefficient table
nl_summary(fit)

# 4. Multilevel R²
nl_r2(fit)

# 5. Model comparison
nl_compare(fit)

# 6. Predictions with CI
pred <- nl_predict(fit, se = TRUE)
nl_plot(pred_df = pred, x = "age", show_ci = TRUE, type = "trajectory")

# 7. Derivatives and turning points
deriv <- nl_derivatives(pred, x = "age")
tp    <- nl_turning_points(deriv, x = "age")
tp$turning_points

nl_plot(deriv_df = deriv, x = "age", type = "slope")
nl_plot(deriv_df = deriv, x = "age", type = "curvature")

# 8. Cluster heterogeneity
nl_het(fit)
```

---

## Changelog

### v0.2.0 (2026)
- **NEW** Two-way and nested clustering (`nested` argument)
- **NEW** CI for `glmerMod` via delta method and parametric bootstrap
- **NEW** Automatic df / knot selection (`df = "auto"`, `nl_knots()`)
- **NEW** Multilevel R² decomposition — marginal, conditional, level-specific (`nl_r2()`)
- **NEW** First and second derivatives with CI (`nl_derivatives()`)
- **NEW** Turning points, inflection regions, slope regions (`nl_turning_points()`)
- **NEW** Built-in model comparison workflow (`nl_compare()`)
- **NEW** Cluster heterogeneity analysis with LRT (`nl_het()`)
- **NEW** B-spline basis (`method = "bs"`, `bs_degree`)
- **NEW** Random spline slopes (`random_slope = TRUE`)
- **ENHANCED** `nl_plot()` now supports `type = "slope"`, `"curvature"`, `"combo"`
- **ENHANCED** `print.nl_fit` lists all v2 postestimation functions

### v0.1.0 (2026-02-27)
- Initial release: `nl_fit`, `nl_predict`, `nl_plot`, `nl_summary`, `nl_icc`

---

## Citation

Hait, S. (2026). *MultiSpline: Spline-Based Nonlinear Modeling for Multilevel
and Longitudinal Data* (R package version 0.2.0).
https://github.com/haitsubi/MultiSpline

---

## References

- Nakagawa, S., & Schielzeth, H. (2013). A general and simple method for
  obtaining R² from generalized linear mixed-effects models. *Methods in
  Ecology and Evolution*, 4(2), 133–142.
- Nakagawa, S., Johnson, P. C. D., & Schielzeth, H. (2017). The coefficient
  of determination R² from generalized linear mixed-effects models revisited
  and expanded. *Journal of the Royal Society Interface*, 14(134).
- Rights, J. D., & Sterba, S. K. (2019). Quantifying explained variance in
  multilevel models. *Psychological Methods*, 24(3), 309–338.
- Wood, S. N. (2017). *Generalized Additive Models: An Introduction with R*
  (2nd ed.). Chapman & Hall/CRC.
- Bates, D., Mächler, M., Bolker, B., & Walker, S. (2015). Fitting linear
  mixed-effects models using lme4. *Journal of Statistical Software*, 67(1).

#' Fit a nonlinear (spline or GAM) single-level or multilevel model
#'
#' @description
#' Fits a nonlinear regression model for an outcome \code{y} with a focal
#' predictor \code{x}, modeled using natural cubic splines (\code{"ns"}),
#' B-splines (\code{"bs"}), or GAM smooths (\code{"gam"}).
#'
#' Version 2 additions: two-way and nested clustering via the
#' \code{nested} argument; random spline slopes via \code{random_slope};
#' B-spline basis via \code{method = "bs"}; automatic df selection via
#' \code{df = "auto"}.
#'
#' @param data A data frame (often long format for longitudinal data).
#' @param y Outcome variable name (string).
#' @param x Focal nonlinear predictor name (string). Must be numeric.
#' @param time Optional time variable name (string).
#' @param cluster Optional character vector of grouping variable name(s) for
#'   random effects, e.g. \code{NULL}, \code{"id"}, or
#'   \code{c("id", "school")} for two-way clustering.
#' @param nested Logical; only used when \code{length(cluster) == 2}. If
#'   \code{TRUE}, uses a nested specification \code{(1 | g1/g2)}; if
#'   \code{FALSE} (default), uses cross-classified
#'   \code{(1 | g1) + (1 | g2)}.
#' @param controls Optional character vector of additional covariate names
#'   to include as linear fixed effects.
#' @param method Spline basis to use: \code{"ns"} (natural cubic spline,
#'   default), \code{"bs"} (B-spline), or \code{"gam"} (GAM smooth via
#'   \code{mgcv::gam()}). Multilevel fits require \code{"ns"} or \code{"bs"}.
#' @param df Degrees of freedom for the spline basis. Supply a single integer
#'   greater than or equal to 1, or the string \code{"auto"} to trigger
#'   automatic selection by information criterion. Default \code{4}.
#' @param df_range Integer vector of candidate df values evaluated when
#'   \code{df = "auto"}. Default \code{2:8}.
#' @param df_criterion Information criterion used for automatic df selection:
#'   \code{"AIC"} (default) or \code{"BIC"}.
#' @param k Basis dimension for \code{mgcv::s()} when \code{method = "gam"}.
#'   Default \code{5}.
#' @param bs_degree Polynomial degree for \code{splines::bs()} when
#'   \code{method = "bs"}. Default \code{3} (cubic).
#' @param random_slope Logical; if \code{TRUE} and \code{cluster} is supplied,
#'   fits random spline slopes in addition to random intercepts. Currently
#'   implemented only for single-level clustering. Default \code{FALSE}.
#' @param family A family object such as \code{stats::gaussian()} (default)
#'   or \code{stats::binomial()}. For multilevel fits, \code{gaussian()} uses
#'   \code{lme4::lmer()} and all other families use \code{lme4::glmer()}.
#' @param ... Additional arguments passed to the underlying fitting function
#'   (\code{stats::lm()}, \code{mgcv::gam()}, \code{lme4::lmer()}, or
#'   \code{lme4::glmer()}).
#'
#' @return An object of class \code{"nl_fit"} (a named list). It contains
#'   the fitted \code{model} object, the \code{method}, variable names
#'   (\code{y}, \code{x}, \code{time}, \code{cluster}, \code{controls}),
#'   spline settings (\code{df}, \code{df_selected}, \code{df_search},
#'   \code{k}, \code{bs_degree}), flags (\code{nested},
#'   \code{random_slope}), the \code{family}, the model \code{formula},
#'   the \code{call}, and metadata used for prediction (\code{x_info},
#'   \code{levels_info}, \code{control_defaults}).
#'
#' @seealso \code{\link{nl_predict}}, \code{\link{nl_derivatives}},
#'   \code{\link{nl_compare}}, \code{\link{nl_r2}}, \code{\link{nl_knots}}
#'
#' @examples
#' \dontrun{
#' # Single-level natural spline with automatic df selection
#' fit <- nl_fit(data = mydata, y = "score", x = "age", df = "auto")
#'
#' # Two-way cross-classified clustering
#' fit2 <- nl_fit(
#'   data    = mydata,
#'   y       = "score",
#'   x       = "age",
#'   cluster = c("student_id", "school_id"),
#'   nested  = FALSE
#' )
#'
#' # Nested clustering (students within schools)
#' fit3 <- nl_fit(
#'   data    = mydata,
#'   y       = "score",
#'   x       = "age",
#'   cluster = c("student_id", "school_id"),
#'   nested  = TRUE
#' )
#'
#' # Random spline slopes
#' fit4 <- nl_fit(
#'   data         = mydata,
#'   y            = "score",
#'   x            = "age",
#'   cluster      = "id",
#'   random_slope = TRUE
#' )
#' }
#'
#' @export
nl_fit <- function(
    data,
    y,
    x,
    time         = NULL,
    cluster      = NULL,
    nested       = FALSE,
    controls     = NULL,
    method       = c("ns", "bs", "gam"),
    df           = 4,
    df_range     = 2:8,
    df_criterion = c("AIC", "BIC"),
    k            = 5,
    bs_degree    = 3,
    random_slope = FALSE,
    family       = stats::gaussian(),
    ...
) {

  # ---------------------------------------------------------------------------
  # 0) Validate inputs
  # ---------------------------------------------------------------------------
  if (!is.data.frame(data))
    stop("`data` must be a data.frame.", call. = FALSE)

  .chk1 <- function(z, nm) {
    if (!is.character(z) || length(z) != 1L || !nzchar(z))
      stop("`", nm, "` must be a single non-empty string.", call. = FALSE)
  }
  .chk1(y, "y")
  .chk1(x, "x")

  method       <- match.arg(method)
  df_criterion <- match.arg(df_criterion)

  if (!is.null(time)) .chk1(time, "time")

  if (!is.null(cluster)) {
    if (!is.character(cluster) || length(cluster) < 1L)
      stop("`cluster` must be NULL or a character vector.", call. = FALSE)
    if (any(!nzchar(cluster)))
      stop("`cluster` contains empty strings.", call. = FALSE)
    cluster <- unique(cluster)
    if (length(cluster) > 2L)
      stop("MultiSpline v2 supports up to two clustering levels.", call. = FALSE)
  }

  if (!is.null(controls)) {
    if (!is.character(controls))
      stop("`controls` must be NULL or a character vector.", call. = FALSE)
    controls <- unique(controls[nzchar(controls)])
    if (length(controls) == 0L) controls <- NULL
  }

  # df can be a number or "auto"
  auto_df <- FALSE
  if (is.character(df) && length(df) == 1L && tolower(df) == "auto") {
    auto_df <- TRUE
    df      <- 4L   # placeholder; replaced after search
  } else {
    if (!is.numeric(df) || length(df) != 1L || !is.finite(df) || df < 1)
      stop("`df` must be a single numeric value >= 1, or \"auto\".",
           call. = FALSE)
    df <- as.integer(df)
  }

  if (!is.numeric(k) || length(k) != 1L || !is.finite(k) || k < 3)
    stop("`k` must be a single numeric value >= 3.", call. = FALSE)

  if (!is.integer(bs_degree) && !is.numeric(bs_degree))
    stop("`bs_degree` must be a positive integer.", call. = FALSE)
  bs_degree <- as.integer(bs_degree)

  if (!is.list(family) || is.null(family$family))
    stop("`family` must be a valid family object.", call. = FALSE)

  # ---------------------------------------------------------------------------
  # 1) Check variable existence in data
  # ---------------------------------------------------------------------------
  needed       <- unique(c(y, x, time, controls, cluster))
  missing_vars <- setdiff(needed, names(data))
  if (length(missing_vars) > 0L)
    stop("Variables not in `data`: ",
         paste(missing_vars, collapse = ", "), call. = FALSE)

  if (!is.numeric(data[[x]]))
    stop("`x` must be numeric. Got class: ", class(data[[x]])[1L],
         call. = FALSE)

  x_vec <- data[[x]]
  if (all(!is.finite(x_vec)))
    stop("`x` contains no finite values.", call. = FALSE)

  is_multilevel <- !is.null(cluster) && length(cluster) >= 1L

  # ---------------------------------------------------------------------------
  # 2) Automatic df selection
  # ---------------------------------------------------------------------------
  df_search <- NULL

  if (auto_df && method != "gam") {
    df_cands <- as.integer(df_range)
    df_cands <- df_cands[df_cands >= 1L]
    if (length(df_cands) == 0L)
      stop("`df_range` produced no valid candidate values.", call. = FALSE)

    crit_vals <- vapply(df_cands, function(dfi) {
      tryCatch({
        tmp <- .build_and_fit(
          data = data, y = y, x = x, time = time,
          cluster = cluster, nested = nested,
          controls = controls, method = method, df = dfi,
          k = k, bs_degree = bs_degree,
          random_slope = random_slope, family = family, dots = list(...)
        )
        if (df_criterion == "AIC") stats::AIC(tmp) else stats::BIC(tmp)
      }, error = function(e) NA_real_)
    }, numeric(1L))

    df_search <- data.frame(df = df_cands, criterion = crit_vals,
                            stringsAsFactors = FALSE)
    best_idx  <- which.min(crit_vals)
    if (length(best_idx) == 0L)
      stop("Automatic df selection failed for all candidate values.",
           call. = FALSE)
    df <- df_cands[best_idx]
    message("Auto knot selection: best df = ", df,
            " (", df_criterion, " = ", round(crit_vals[best_idx], 2), ")")
  }

  # ---------------------------------------------------------------------------
  # 3) Fit the selected model
  # ---------------------------------------------------------------------------
  model <- .build_and_fit(
    data = data, y = y, x = x, time = time,
    cluster = cluster, nested = nested,
    controls = controls, method = method, df = df,
    k = k, bs_degree = bs_degree,
    random_slope = random_slope, family = family, dots = list(...)
  )

  form <- tryCatch(stats::formula(model), error = function(e) NULL)

  # ---------------------------------------------------------------------------
  # 4) Store compact training metadata
  # ---------------------------------------------------------------------------
  x_q <- suppressWarnings(
    stats::quantile(x_vec, probs = c(0.01, 0.99), na.rm = TRUE, names = FALSE)
  )
  x_r <- range(x_vec, na.rm = TRUE)
  if (any(!is.finite(x_q))) x_q <- x_r

  x_info <- list(
    q01 = unname(x_q[1L]), q99 = unname(x_q[2L]),
    min = unname(x_r[1L]), max = unname(x_r[2L]),
    n   = sum(is.finite(x_vec))
  )

  levels_info      <- list()
  control_defaults <- list()

  if (!is.null(time) && is.factor(data[[time]]))
    levels_info[[time]] <- levels(data[[time]])

  if (!is.null(controls)) {
    for (cn in controls) {
      vec <- data[[cn]]
      if (is.factor(vec)) {
        levels_info[[cn]] <- levels(vec)
      } else if (is.character(vec)) {
        # Store unique sorted values so newdata can be properly factored
        # for model.matrix — this prevents column-name mismatches in nl_predict
        levels_info[[cn]] <- sort(unique(vec[!is.na(vec)]))
      }
      control_defaults[[cn]] <- if (is.numeric(vec)) {
        mean(vec, na.rm = TRUE)
      } else if (is.factor(vec)) {
        levels(vec)[1L]
      } else {
        vec[which(!is.na(vec))[1L]]
      }
    }
  }

  # ---------------------------------------------------------------------------
  # 5) Return nl_fit object
  # ---------------------------------------------------------------------------
  structure(
    list(
      model            = model,
      method           = method,
      y                = y,
      x                = x,
      time             = time,
      cluster          = if (is_multilevel) cluster else NULL,
      nested           = nested,
      controls         = controls,
      df               = df,
      df_selected      = df,
      df_search        = df_search,
      k                = k,
      bs_degree        = bs_degree,
      random_slope     = random_slope,
      family           = family,
      formula          = form,
      call             = match.call(),
      x_info           = x_info,
      levels_info      = levels_info,
      control_defaults = control_defaults
    ),
    class = "nl_fit"
  )
}


# ============================================================
# Internal: construct formula and call the fitting engine
# ============================================================
.build_and_fit <- function(data, y, x, time, cluster, nested,
                           controls, method, df, k, bs_degree,
                           random_slope, family, dots) {

  is_ml <- !is.null(cluster) && length(cluster) >= 1L

  # Fixed-effects spline term
  main_part <- switch(method,

    ns = {
      sc <- paste0("splines::ns(", x, ", df = ", df, ")")
      if (!is.null(time)) paste0(sc, " * ", time) else sc
    },

    bs = {
      sc <- paste0("splines::bs(", x, ", df = ", df,
                   ", degree = ", bs_degree, ")")
      if (!is.null(time)) paste0(sc, " * ", time) else sc
    },

    gam = {
      if (!is.null(time)) {
        time_is_fac <- is.factor(data[[time]]) || is.character(data[[time]])
        if (time_is_fac)
          paste0("s(", x, ", by = ", time, ", k = ", k, ") + ", time)
        else
          paste0("s(", x, ", k = ", k, ") + ", time,
                 " + ti(", x, ", ", time, ", k = ", k, ")")
      } else {
        paste0("s(", x, ", k = ", k, ")")
      }
    }
  )

  fixed_part <- paste(c(main_part, controls), collapse = " + ")

  if (!is_ml) {
    # ---- Single-level ----
    form_env <- if (method == "gam") asNamespace("mgcv") else parent.frame()
    form     <- stats::as.formula(paste0(y, " ~ ", fixed_part),
                                  env = form_env)
    if (method == "gam") {
      do.call(mgcv::gam,
              c(list(formula = form, data = data, family = family), dots))
    } else {
      do.call(stats::lm, c(list(formula = form, data = data), dots))
    }

  } else {
    # ---- Multilevel ----
    if (method == "gam")
      stop("Multilevel support requires method = 'ns' or 'bs'.", call. = FALSE)

    rand_part <- .build_rand_formula(
      cluster      = cluster,
      nested       = nested,
      x            = x,
      df           = df,
      bs_degree    = bs_degree,
      method       = method,
      random_slope = random_slope
    )

    form <- stats::as.formula(
      paste0(y, " ~ ", fixed_part, " + ", rand_part)
    )

    if (identical(family$family, "gaussian")) {
      do.call(lme4::lmer, c(list(formula = form, data = data), dots))
    } else {
      do.call(lme4::glmer,
              c(list(formula = form, data = data, family = family), dots))
    }
  }
}


# ============================================================
# Internal: build the random-effects formula string
# ============================================================
.build_rand_formula <- function(cluster, nested, x, df, bs_degree,
                                method, random_slope) {

  n_clust <- length(cluster)

  if (!random_slope || n_clust != 1L) {
    # Random intercepts only
    if (n_clust == 1L) {
      rand_part <- paste0("(1 | ", cluster[1L], ")")
    } else {
      rand_part <- if (nested) {
        paste0("(1 | ", cluster[1L], "/", cluster[2L], ")")
      } else {
        paste0("(1 | ", cluster[1L], ") + (1 | ", cluster[2L], ")")
      }
    }
  } else {
    # Random spline slopes (single cluster variable only)
    sc <- if (method == "ns") {
      paste0("splines::ns(", x, ", df = ", df, ")")
    } else {
      paste0("splines::bs(", x, ", df = ", df,
             ", degree = ", bs_degree, ")")
    }
    rand_part <- paste0("(", sc, " | ", cluster[1L], ")")
  }

  rand_part
}

#' Multilevel R-squared decomposition for nl_fit models
#'
#' @description
#' Computes a suite of R-squared statistics for models fitted by
#' \code{\link{nl_fit}}.
#' For single-level models the standard R-squared and adjusted R-squared are
#' returned. For multilevel models (lmerMod / glmerMod) two quantities are
#' reported: the Nakagawa-Schielzeth marginal R-squared (variance explained by
#' fixed effects only) and the conditional R-squared (fixed plus all random
#' effects), together with a level-specific variance partition table analogous
#' to the r2_mlm / Raudenbush-Bryk approach.
#'
#' @details
#' Marginal and conditional R-squared for LMMs follow the Nakagawa and
#' Schielzeth (2013) formulae extended to multiple random effects by Nakagawa,
#' Johnson and Schielzeth (2017). The fixed-effects variance
#' \eqn{\sigma^2_f} is computed as the variance of the linear predictor
#' from fixed effects only (\eqn{\hat{\mu} = X\hat{\beta}}).
#'
#' The level-specific variance partition (r2_mlm-style) decomposes the total
#' modelled variance (\eqn{\sigma^2_f + \sum \sigma^2_j + \sigma^2_\epsilon})
#' to show how much each source contributes, printed as a breakdown table.
#'
#' @param object An \code{nl_fit} object returned by \code{\link{nl_fit}}.
#' @param digits Integer; decimal places for display. Default \code{4}.
#'
#' @return A list of class \code{"nl_r2"} returned invisibly and
#'   pretty-printed automatically. It contains \code{type} (one of
#'   \code{"OLS"}, \code{"GAM"}, \code{"LMM"}, or \code{"GLMM"}),
#'   \code{r2} (a named numeric vector: \code{R2} and \code{R2_adj} for OLS;
#'   \code{R2_dev} for GAM; \code{R2m} and \code{R2c} for LMM/GLMM), and
#'   \code{variance_partition} (a data frame with columns \code{component},
#'   \code{variance}, and \code{proportion} for multilevel models, or
#'   \code{NULL} for single-level models).
#'
#' @references
#' Nakagawa, S., & Schielzeth, H. (2013). A general and simple method for
#' obtaining R-squared from generalized linear mixed-effects models.
#' Methods in Ecology and Evolution, 4(2), 133--142.
#'
#' Nakagawa, S., Johnson, P. C. D., & Schielzeth, H. (2017). The coefficient
#' of determination R-squared and intra-class correlation coefficient from
#' generalized linear mixed-effects models revisited and expanded.
#' Journal of the Royal Society Interface, 14(134), 20170213.
#'
#' Rights, J. D., & Sterba, S. K. (2019). Quantifying explained variance in
#' multilevel models: An integrative framework for defining R-squared measures.
#' Psychological Methods, 24(3), 309--338.
#'
#' @seealso \code{\link{nl_fit}}, \code{\link{nl_icc}}
#'
#' @export
nl_r2 <- function(object, digits = 4L) {

  if (!inherits(object, "nl_fit"))
    stop("`object` must be an 'nl_fit' object.", call. = FALSE)

  mod <- object$model

  # ===========================================================================
  # OLS / single-level lm
  # ===========================================================================
  if (inherits(mod, "lm") && !inherits(mod, "gam")) {
    s   <- summary(mod)
    r2  <- c(R2 = s$r.squared, R2_adj = s$adj.r.squared)
    out <- structure(
      list(type = "OLS", r2 = r2, variance_partition = NULL),
      class = "nl_r2"
    )
    print(out, digits = digits)
    return(invisible(out))
  }

  # ===========================================================================
  # GAM -- deviance-based R-squared
  # ===========================================================================
  if (inherits(mod, "gam")) {
    s   <- summary(mod)
    r2  <- c(R2_dev = s$dev.expl)
    out <- structure(
      list(type = "GAM", r2 = r2, variance_partition = NULL),
      class = "nl_r2"
    )
    print(out, digits = digits)
    return(invisible(out))
  }

  # ===========================================================================
  # LMM (lmerMod)
  # ===========================================================================
  if (inherits(mod, "lmerMod")) {

    vc   <- lme4::VarCorr(mod)
    vcdf <- as.data.frame(vc)

    # Random-intercept variances per grouping factor
    re_rows <- vcdf[
      which(!is.na(vcdf$grp) &
            vcdf$var1 == "(Intercept)" &
            is.na(vcdf$var2) &
            !is.na(vcdf$vcov)), , drop = FALSE
    ]

    if (!nrow(re_rows))
      stop(
        "No random intercept variances found. Is this truly a multilevel model?",
        call. = FALSE
      )

    re_var   <- stats::setNames(as.numeric(re_rows$vcov),
                                as.character(re_rows$grp))
    # Defensive: drop any entries where name or value is NA
    re_var   <- re_var[!is.na(names(re_var)) & !is.na(re_var)]
    sigma_e2 <- as.numeric(attr(vc, "sc"))^2   # residual variance, unname

    # Fixed-effects variance: var(X * beta)
    ff_nobars <- tryCatch(
      if (requireNamespace("reformulas", quietly = TRUE))
        reformulas::nobars(stats::formula(mod))
      else
        lme4::nobars(stats::formula(mod)),
      error = function(e) NULL
    )

    sigma_f2 <- if (!is.null(ff_nobars)) {
      mf <- stats::model.frame(mod)
      tt <- stats::delete.response(stats::terms(ff_nobars))
      X  <- tryCatch(stats::model.matrix(tt, mf), error = function(e) NULL)
      if (!is.null(X)) {
        beta   <- lme4::fixef(mod)
        common <- intersect(colnames(X), names(beta))
        Xb     <- X[, common, drop = FALSE] %*% beta[common]
        stats::var(as.numeric(Xb))
      } else NA_real_
    } else NA_real_

    total_re  <- sum(re_var, na.rm = TRUE)
    total_var <- sigma_f2 + total_re + sigma_e2

    R2m <- sigma_f2 / total_var
    R2c <- (sigma_f2 + total_re) / total_var
    r2  <- c(R2m = R2m, R2c = R2c)

    # Variance partition (r2_mlm style)
    parts <- c(re_var, Residual = sigma_e2)
    vp_df <- data.frame(
      component  = c("Fixed effects", names(parts)),
      variance   = c(sigma_f2, as.numeric(parts)),
      proportion = c(sigma_f2, as.numeric(parts)) / total_var,
      stringsAsFactors = FALSE
    )

    out <- structure(
      list(type = "LMM", r2 = r2, variance_partition = vp_df),
      class = "nl_r2"
    )
    print(out, digits = digits)
    return(invisible(out))
  }

  # ===========================================================================
  # GLMM (glmerMod)
  # ===========================================================================
  if (inherits(mod, "glmerMod")) {

    fam  <- stats::family(mod)
    vc   <- lme4::VarCorr(mod)
    vcdf <- as.data.frame(vc)

    re_rows <- vcdf[
      which(!is.na(vcdf$grp) &
            vcdf$var1 == "(Intercept)" &
            is.na(vcdf$var2) &
            !is.na(vcdf$vcov)), , drop = FALSE
    ]

    re_var   <- stats::setNames(as.numeric(re_rows$vcov),
                                as.character(re_rows$grp))
    re_var   <- re_var[!is.na(names(re_var)) & !is.na(re_var)]
    total_re <- sum(re_var, na.rm = TRUE)

    # Fixed-effects variance
    ff_nobars <- tryCatch(
      if (requireNamespace("reformulas", quietly = TRUE))
        reformulas::nobars(stats::formula(mod))
      else
        lme4::nobars(stats::formula(mod)),
      error = function(e) NULL
    )

    sigma_f2 <- if (!is.null(ff_nobars)) {
      mf <- stats::model.frame(mod)
      tt <- stats::delete.response(stats::terms(ff_nobars))
      X  <- tryCatch(stats::model.matrix(tt, mf), error = function(e) NULL)
      if (!is.null(X)) {
        beta   <- lme4::fixef(mod)
        common <- intersect(colnames(X), names(beta))
        Xb     <- X[, common, drop = FALSE] %*% beta[common]
        stats::var(as.numeric(Xb))
      } else NA_real_
    } else NA_real_

    # Distribution-specific variance (link-scale approximation)
    sigma_d2 <- switch(fam$family,
      binomial = if (fam$link == "logit") (pi^2) / 3 else 1,
      poisson  = {
        mu <- stats::fitted(mod)
        log(1 + 1 / mean(mu, na.rm = TRUE))
      },
      NA_real_
    )

    total_var <- sigma_f2 + total_re + sigma_d2
    R2m <- sigma_f2 / total_var
    R2c <- (sigma_f2 + total_re) / total_var
    r2  <- c(R2m = R2m, R2c = R2c)

    parts <- c(re_var, Distribution = sigma_d2)
    vp_df <- data.frame(
      component  = c("Fixed effects", names(parts)),
      variance   = c(sigma_f2, as.numeric(parts)),
      proportion = c(sigma_f2, as.numeric(parts)) / total_var,
      stringsAsFactors = FALSE
    )

    out <- structure(
      list(type = "GLMM", r2 = r2, variance_partition = vp_df),
      class = "nl_r2"
    )
    print(out, digits = digits)
    return(invisible(out))
  }

  stop("Unknown model class inside nl_fit.", call. = FALSE)
}


#' @export
print.nl_r2 <- function(x, digits = 4L, ...) {

  cat("\n=== MultiSpline R-squared Decomposition ===\n")
  cat("Model type:", x$type, "\n\n")

  if (x$type == "OLS") {
    cat(sprintf("  R-squared          = %.4f\n", x$r2["R2"]))
    cat(sprintf("  Adjusted R-squared = %.4f\n", x$r2["R2_adj"]))

  } else if (x$type == "GAM") {
    cat(sprintf("  Deviance explained (R-squared) = %.4f\n", x$r2["R2_dev"]))

  } else {
    # LMM / GLMM
    cat(sprintf("  Marginal  R2m = %.4f   (fixed effects only)\n",
                x$r2["R2m"]))
    cat(sprintf("  Conditional R2c = %.4f  (fixed + all random effects)\n",
                x$r2["R2c"]))

    if (!is.null(x$variance_partition) && nrow(x$variance_partition) > 0L) {
      cat("\nVariance Partition (r2_mlm-style):\n")
      vp <- x$variance_partition
      vp_disp <- data.frame(
        Component  = vp$component,
        Variance   = round(vp$variance,   digits),
        Proportion = round(vp$proportion, digits),
        stringsAsFactors = FALSE
      )
      print(vp_disp, row.names = FALSE)
    }
  }

  cat("\n")
  invisible(x)
}

#' Plot predictions and derivatives from nl_fit models
#'
#' @description
#' Visualises results from \code{\link{nl_predict}} or
#' \code{\link{nl_derivatives}}.  The \code{type} argument selects the plot:
#' \itemize{
#'   \item \code{"trajectory"}: predicted outcome vs \code{x} with optional
#'     confidence ribbon (v1 behaviour, now default).
#'   \item \code{"slope"}: first derivative (marginal effect) vs \code{x}.
#'   \item \code{"curvature"}: second derivative vs \code{x}.
#'   \item \code{"combo"}: trajectory + slope panels side by side.
#' }
#'
#' @param pred_df A data frame from \code{\link{nl_predict}} (for
#'   \code{type = "trajectory"} or \code{"combo"}).
#' @param deriv_df A data frame from \code{\link{nl_derivatives}} (for
#'   \code{type = "slope"}, \code{"curvature"}, or \code{"combo"}).
#' @param x Character; name of the focal predictor column in \code{pred_df} /
#'   \code{deriv_df}.
#' @param time Optional character; name of the time column. Auto-detected if
#'   \code{NULL}.
#' @param type Plot type: \code{"trajectory"} (default), \code{"slope"},
#'   \code{"curvature"}, or \code{"combo"}.
#' @param show_ci Logical; add confidence ribbons. Default \code{TRUE}.
#' @param ci_level Numeric; confidence level when deriving CI from
#'   \code{se.fit}. Default \code{0.95}.
#' @param show_turning_points Logical; overlay turning-point markers on the
#'   trajectory plot. Default \code{FALSE}.
#' @param turning_points Data frame from \code{\link{nl_turning_points}} (only
#'   used when \code{show_turning_points = TRUE}).
#' @param zero_line Logical; for slope / curvature plots, draw a dashed
#'   horizontal line at zero. Default \code{TRUE}.
#' @param y_lab Y-axis label.
#' @param x_lab X-axis label. Defaults to \code{x}.
#' @param title Optional plot title.
#' @param legend_title Optional legend title.
#'
#' @return A \code{ggplot} object (or a combined plot list for \code{"combo"}).
#'
#' @seealso \code{\link{nl_predict}}, \code{\link{nl_derivatives}},
#'   \code{\link{nl_turning_points}}
#'
#' @importFrom rlang .data
#' @export
nl_plot <- function(
    pred_df          = NULL,
    deriv_df         = NULL,
    x,
    time             = NULL,
    type             = c("trajectory", "slope", "curvature", "combo"),
    show_ci          = TRUE,
    ci_level         = 0.95,
    show_turning_points = FALSE,
    turning_points   = NULL,
    zero_line        = TRUE,
    y_lab            = NULL,
    x_lab            = NULL,
    title            = NULL,
    legend_title     = NULL
) {

  type <- match.arg(type)

  # Auto-detect time
  .detect_time <- function(df) {
    if (is.null(df)) return(NULL)
    candidates <- c("TimePoint","time","Time","wave","Wave","occasion","Occasion")
    hit <- candidates[candidates %in% names(df)]
    if (length(hit)) hit[1L] else NULL
  }
  if (is.null(time)) {
    time <- .detect_time(pred_df) %||% .detect_time(deriv_df)
  }

  if (is.null(x_lab)) x_lab <- x
  if (is.null(legend_title)) legend_title <- time

  # Helper: build CI from se.fit if lwr/upr absent
  .ensure_ci <- function(df) {
    if (!is.data.frame(df)) return(df)
    has_lwr <- all(c("lwr","upr") %in% names(df))
    has_se  <- "se.fit" %in% names(df) && any(!is.na(df$se.fit))
    if (!has_lwr && has_se) {
      crit    <- stats::qnorm(1 - (1 - ci_level) / 2)
      df$lwr  <- df$fit - crit * df$se.fit
      df$upr  <- df$fit + crit * df$se.fit
    }
    df
  }

  # ---- Shared plot base ----
  .base_aes <- function(df, y_col) {
    if (!is.null(time) && time %in% names(df)) {
      df <- df[order(df[[time]], df[[x]]), , drop = FALSE]
    } else {
      df <- df[order(df[[x]]), , drop = FALSE]
    }
    if (!is.null(time) && time %in% names(df)) {
      ggplot2::ggplot(df, ggplot2::aes(
        x = .data[[x]], y = .data[[y_col]],
        color = .data[[time]], group = .data[[time]]
      ))
    } else {
      ggplot2::ggplot(df, ggplot2::aes(x = .data[[x]], y = .data[[y_col]]))
    }
  }

  .add_ribbon <- function(p, df, lo_col, hi_col) {
    if (!all(c(lo_col, hi_col) %in% names(df))) return(p)
    if (!is.null(time) && time %in% names(df)) {
      p + ggplot2::geom_ribbon(
        ggplot2::aes(ymin = .data[[lo_col]], ymax = .data[[hi_col]],
                     fill = .data[[time]], group = .data[[time]]),
        alpha = 0.14, color = NA)
    } else {
      p + ggplot2::geom_ribbon(
        ggplot2::aes(ymin = .data[[lo_col]], ymax = .data[[hi_col]]),
        alpha = 0.18, color = NA)
    }
  }

  .theme <- function() {
    list(
      ggplot2::geom_line(linewidth = 0.9),
      ggplot2::theme_minimal(base_size = 12),
      ggplot2::theme(
        legend.position  = if (!is.null(time)) "right" else "none",
        plot.title       = ggplot2::element_text(face = "bold")
      )
    )
  }

  # =========================================================
  # TRAJECTORY plot
  # =========================================================
  .plot_traj <- function() {
    if (is.null(pred_df)) stop("`pred_df` required for type='trajectory'.", call. = FALSE)
    df <- .ensure_ci(as.data.frame(pred_df))
    p  <- .base_aes(df, "fit")
    if (isTRUE(show_ci) && all(c("lwr","upr") %in% names(df)))
      p <- .add_ribbon(p, df, "lwr", "upr")
    p <- p + .theme() +
      ggplot2::labs(x = x_lab,
                    y = if (!is.null(y_lab)) y_lab else "Predicted outcome",
                    title = title, color = legend_title, fill = legend_title)

    if (isTRUE(show_turning_points) && !is.null(turning_points) &&
        nrow(turning_points) > 0L && x %in% names(turning_points)) {
      p <- p +
        ggplot2::geom_vline(
          data    = turning_points,
          mapping = ggplot2::aes(xintercept = .data[[x]],
                                  linetype    = .data[["type"]]),
          color = "grey30", linewidth = 0.6
        ) +
        ggplot2::scale_linetype_manual(
          values = c(maximum = "dashed", minimum = "dotted", saddle = "dotdash")
        )
    }
    p
  }

  # =========================================================
  # SLOPE (first derivative) plot
  # =========================================================
  .plot_slope <- function() {
    if (is.null(deriv_df)) stop("`deriv_df` required for type='slope'.", call. = FALSE)
    df <- as.data.frame(deriv_df)
    p  <- .base_aes(df, "d1")
    if (isTRUE(show_ci) && all(c("d1_lwr","d1_upr") %in% names(df)))
      p <- .add_ribbon(p, df, "d1_lwr", "d1_upr")
    if (isTRUE(zero_line))
      p <- p + ggplot2::geom_hline(yintercept = 0, linetype = "dashed",
                                    color = "grey50", linewidth = 0.5)
    p <- p + .theme() +
      ggplot2::labs(
        x     = x_lab,
        y     = if (!is.null(y_lab)) y_lab else paste0("dy / d(", x, ")"),
        title = title %||% paste0("Marginal Effect (First Derivative)  of ", x),
        color = legend_title, fill = legend_title
      )
    p
  }

  # =========================================================
  # CURVATURE (second derivative) plot
  # =========================================================
  .plot_curv <- function() {
    if (is.null(deriv_df)) stop("`deriv_df` required for type='curvature'.", call. = FALSE)
    df <- as.data.frame(deriv_df)
    if (!"d2" %in% names(df)) stop("`deriv_df` must contain column 'd2'.", call. = FALSE)
    p  <- .base_aes(df, "d2")
    if (isTRUE(show_ci) && all(c("d2_lwr","d2_upr") %in% names(df)))
      p <- .add_ribbon(p, df, "d2_lwr", "d2_upr")
    if (isTRUE(zero_line))
      p <- p + ggplot2::geom_hline(yintercept = 0, linetype = "dashed",
                                    color = "grey50", linewidth = 0.5)
    p <- p + .theme() +
      ggplot2::labs(
        x     = x_lab,
        y     = if (!is.null(y_lab)) y_lab else paste0("d2(y) / d(", x, ")^2"),
        title = title %||% paste0("Curvature (Second Derivative) of ", x),
        color = legend_title, fill = legend_title
      )
    p
  }

  switch(type,
    trajectory = .plot_traj(),
    slope      = .plot_slope(),
    curvature  = .plot_curv(),
    combo      = {
      p1 <- .plot_traj()
      p2 <- .plot_slope()
      list(trajectory = p1, slope = p2)
    }
  )
}

#' Built-in model comparison workflow
#'
#' @description
#' Compares a nonlinear spline model against simpler alternatives—a linear
#' model and optional polynomial terms—to help researchers justify the
#' spline approach over simpler specifications.
#'
#' For each model, the function reports:
#' \itemize{
#'   \item AIC and BIC
#'   \item Log-likelihood (and likelihood-ratio test versus the linear model
#'     where available)
#'   \item Number of parameters (df)
#'   \item Residual variance / deviance
#' }
#'
#' @param object An \code{nl_fit} object (the spline model to compare against).
#' @param polynomial_degrees Integer vector of polynomial degrees to include
#'   as additional comparators. Default \code{c(2L, 3L)}. Set to
#'   \code{integer(0)} to skip.
#' @param digits Integer; decimal places for the output table. Default \code{3}.
#' @param return_models Logical; if \code{TRUE}, also returns the fitted
#'   comparison model objects. Default \code{FALSE}.
#'
#' @return A list (invisibly) of class \code{"nl_compare"} with:
#'   \describe{
#'     \item{\code{table}}{A data frame with the comparison statistics.}
#'     \item{\code{models}}{Named list of fitted models (when
#'       \code{return_models = TRUE}).}
#'     \item{\code{best}}{Name of the model with the lowest AIC.}
#'   }
#'   The table is pretty-printed automatically.
#'
#' @examples
#' \dontrun{
#' fit <- nl_fit(data = mydata, y = "score", x = "age", df = 4)
#' nl_compare(fit)
#' nl_compare(fit, polynomial_degrees = c(2, 3, 4))
#' }
#'
#' @seealso \code{\link{nl_fit}}, \code{\link{nl_knots}}
#'
#' @export
nl_compare <- function(
    object,
    polynomial_degrees = c(2L, 3L),
    digits             = 3L,
    return_models      = FALSE
) {

  if (!inherits(object, "nl_fit"))
    stop("`object` must be an 'nl_fit' object.", call. = FALSE)

  data     <- tryCatch(stats::model.frame(object$model), error = function(e) NULL)
  if (is.null(data))
    stop("Cannot extract model frame from fitted model.", call. = FALSE)

  y        <- object$y
  x        <- object$x
  controls <- object$controls
  cluster  <- object$cluster
  nested   <- object$nested
  family   <- object$family
  method   <- object$method
  df       <- object$df
  k        <- object$k
  bs_deg   <- object$bs_degree

  ctrl_part <- if (!is.null(controls)) paste(controls, collapse = " + ") else NULL

  # Helper: fit a model and extract fit stats
  .fit_stats <- function(nm, mod) {
    tryCatch({
      aic <- stats::AIC(mod)
      bic <- stats::BIC(mod)
      ll  <- as.numeric(stats::logLik(mod))
      npar<- attr(stats::logLik(mod), "df")
      dev <- if (inherits(mod, "lm")) {
        stats::deviance(mod)
      } else if (inherits(mod, "lmerMod")) {
        # Use ML deviance to avoid the REML deprecation warning
        tryCatch(suppressWarnings(stats::deviance(mod, REML = FALSE)),
                 error = function(e) NA_real_)
      } else {
        tryCatch(stats::deviance(mod), error = function(e) NA_real_)
      }
      data.frame(Model = nm, AIC = aic, BIC = bic, LogLik = ll,
                 npar = npar, Deviance = dev,
                 LRT_vs_linear = NA_real_, LRT_p = NA_real_,
                 stringsAsFactors = FALSE)
    }, error = function(e) {
      message("Could not fit model '", nm, "': ", conditionMessage(e))
      NULL
    })
  }

  # --- 1. Linear model ---
  lin_rhs  <- paste(c(x, controls), collapse = " + ")
  lin_form <- if (!is.null(cluster)) {
    rand <- if (length(cluster) == 1L) {
      paste0("(1 | ", cluster, ")")
    } else if (nested) {
      paste0("(1 | ", cluster[1L], "/", cluster[2L], ")")
    } else {
      paste0("(1 | ", cluster[1L], ") + (1 | ", cluster[2L], ")")
    }
    stats::as.formula(paste0(y, " ~ ", lin_rhs, " + ", rand))
  } else {
    stats::as.formula(paste0(y, " ~ ", lin_rhs))
  }

  orig_data <- tryCatch(
    eval(object$call$data, envir = parent.frame()),
    error = function(e) NULL
  )
  fit_data <- if (!is.null(orig_data)) orig_data else data

  lin_mod <- tryCatch({
    if (!is.null(cluster)) {
      if (identical(family$family, "gaussian"))
        lme4::lmer(lin_form, data = fit_data)
      else
        lme4::glmer(lin_form, data = fit_data, family = family)
    } else if (method == "gam") {
      stats::lm(lin_form, data = fit_data)
    } else {
      stats::lm(lin_form, data = fit_data)
    }
  }, error = function(e) { message("Linear model failed: ", e$message); NULL })

  model_list <- list()
  if (!is.null(lin_mod)) model_list[["Linear"]] <- lin_mod

  # --- 2. Polynomial models ---
  if (length(polynomial_degrees) > 0L) {
    for (deg in as.integer(polynomial_degrees)) {
      poly_main <- paste0("poly(", x, ", ", deg, ")")
      poly_rhs  <- paste(c(poly_main, controls), collapse = " + ")
      poly_form <- if (!is.null(cluster)) {
        rand <- if (length(cluster) == 1L) {
          paste0("(1 | ", cluster, ")")
        } else if (nested) {
          paste0("(1 | ", cluster[1L], "/", cluster[2L], ")")
        } else {
          paste0("(1 | ", cluster[1L], ") + (1 | ", cluster[2L], ")")
        }
        stats::as.formula(paste0(y, " ~ ", poly_rhs, " + ", rand))
      } else {
        stats::as.formula(paste0(y, " ~ ", poly_rhs))
      }
      nm <- paste0("Poly(", deg, ")")
      pm <- tryCatch({
        if (!is.null(cluster)) {
          if (identical(family$family, "gaussian"))
            lme4::lmer(poly_form, data = fit_data)
          else
            lme4::glmer(poly_form, data = fit_data, family = family)
        } else {
          stats::lm(poly_form, data = fit_data)
        }
      }, error = function(e) { message(nm, " failed: ", e$message); NULL })
      if (!is.null(pm)) model_list[[nm]] <- pm
    }
  }

  # --- 3. Spline model (the nl_fit object) ---
  model_list[["Spline"]] <- object$model

  # --- 4. Collect fit statistics ---
  rows <- lapply(names(model_list), function(nm) {
    .fit_stats(nm, model_list[[nm]])
  })
  rows <- Filter(Negate(is.null), rows)
  tbl  <- do.call(rbind, rows)

  # --- 5. LRT vs linear ---
  if (!is.null(lin_mod) && nrow(tbl) > 1L) {
    lin_ll   <- as.numeric(stats::logLik(lin_mod))
    lin_npar <- attr(stats::logLik(lin_mod), "df")
    for (i in seq_len(nrow(tbl))) {
      nm_i <- tbl$Model[i]
      if (nm_i == "Linear") next
      mod_i <- model_list[[nm_i]]
      if (is.null(mod_i)) next
      ll_i   <- tbl$LogLik[i]
      npar_i <- tbl$npar[i]
      delta_df <- npar_i - lin_npar
      if (is.finite(ll_i) && is.finite(lin_ll) && delta_df > 0) {
        lrt_stat <- 2 * (ll_i - lin_ll)
        tbl$LRT_vs_linear[i] <- round(lrt_stat, digits)
        tbl$LRT_p[i]         <- stats::pchisq(lrt_stat, df = delta_df,
                                               lower.tail = FALSE)
      }
    }
  }

  # Nicely round
  for (col in c("AIC","BIC","LogLik","Deviance","LRT_vs_linear")) {
    if (col %in% names(tbl)) tbl[[col]] <- round(tbl[[col]], digits)
  }
  tbl$LRT_p <- ifelse(is.na(tbl$LRT_p), NA_character_,
                      format.pval(tbl$LRT_p, digits = digits, eps = 0.001))

  best <- if (nrow(tbl) > 0L) tbl$Model[which.min(tbl$AIC)] else NA_character_

  out <- structure(
    list(
      table  = tbl,
      models = if (return_models) model_list else NULL,
      best   = best
    ),
    class = "nl_compare"
  )

  print(out)
  invisible(out)
}


#' @export
print.nl_compare <- function(x, ...) {
  cat("\n=== MultiSpline Model Comparison ===\n\n")
  print(x$table, row.names = FALSE, ...)
  cat("\n  Best model by AIC:", x$best, "\n\n")
  invisible(x)
}

#' Automatic knot / degrees-of-freedom selection for spline models
#'
#' @description
#' Performs a grid search over candidate degrees of freedom (df) for a
#' natural cubic spline (\code{"ns"}) or B-spline (\code{"bs"}) model and
#' selects the best value by an information criterion (AIC or BIC). A
#' diagnostic plot of the criterion against df is returned.
#'
#' This function is called internally by \code{\link{nl_fit}} when
#' \code{df = "auto"}, but it can also be called directly for exploration
#' before fitting the final model.
#'
#' @param data A data frame.
#' @param y Outcome variable name (string).
#' @param x Focal predictor name (string).
#' @param time Optional time variable name.
#' @param cluster Optional character vector of cluster variable names.
#' @param nested Logical; nested clustering. Default \code{FALSE}.
#' @param controls Optional character vector of control variable names.
#' @param method Either \code{"ns"} (default) or \code{"bs"}.
#' @param df_range Integer vector of candidate df values. Default \code{2:10}.
#' @param criterion Either \code{"AIC"} (default) or \code{"BIC"}.
#' @param family A family object. Default \code{stats::gaussian()}.
#' @param plot Logical; if \code{TRUE}, prints a diagnostic plot of criterion
#'   vs df. Default \code{TRUE}.
#' @param ... Additional arguments passed to the underlying fitter.
#'
#' @return A list with:
#'   \describe{
#'     \item{\code{best_df}}{The df value with the lowest criterion value.}
#'     \item{\code{search_table}}{Data frame with columns \code{df} and
#'       \code{criterion}.}
#'     \item{\code{criterion}}{The criterion used (\code{"AIC"} or
#'       \code{"BIC"}).}
#'     \item{\code{plot}}{A \code{ggplot} object (if \code{plot = TRUE}).}
#'   }
#'
#' @seealso \code{\link{nl_fit}}, \code{\link{nl_compare}}
#'
#' @export
nl_knots <- function(
    data,
    y,
    x,
    time      = NULL,
    cluster   = NULL,
    nested    = FALSE,
    controls  = NULL,
    method    = c("ns", "bs"),
    df_range  = 2:10,
    criterion = c("AIC", "BIC"),
    family    = stats::gaussian(),
    plot      = TRUE,
    ...
) {

  method    <- match.arg(method)
  criterion <- match.arg(criterion)
  df_cands  <- as.integer(df_range)
  df_cands  <- df_cands[df_cands >= 1L]
  if (length(df_cands) == 0L) stop("`df_range` produced no valid values.", call. = FALSE)

  dots <- list(...)

  crit_vals <- vapply(df_cands, function(dfi) {
    tryCatch({
      mod <- .build_and_fit(
        data = data, y = y, x = x, time = time,
        cluster = cluster, nested = nested,
        controls = controls, method = method, df = dfi,
        k = 5L, bs_degree = 3L,
        random_slope = FALSE, family = family, dots = dots
      )
      if (criterion == "AIC") stats::AIC(mod) else stats::BIC(mod)
    }, error = function(e) NA_real_)
  }, numeric(1L))

  tbl     <- data.frame(df = df_cands, criterion = crit_vals,
                        stringsAsFactors = FALSE)
  best_df <- df_cands[which.min(crit_vals)]

  p <- NULL
  if (plot) {
    p <- ggplot2::ggplot(
      tbl[is.finite(tbl$criterion), ],
      ggplot2::aes(x = .data[["df"]], y = .data[["criterion"]])
    ) +
      ggplot2::geom_line(color = "#2171B5", linewidth = 0.9) +
      ggplot2::geom_point(size = 2.5, color = "#2171B5") +
      ggplot2::geom_vline(xintercept = best_df, linetype = "dashed",
                          color = "#D7191C", linewidth = 0.8) +
      ggplot2::annotate("text", x = best_df, y = max(tbl$criterion, na.rm = TRUE),
                        label = paste0("Best df = ", best_df),
                        hjust = -0.1, vjust = 1.2, color = "#D7191C", size = 3.5) +
      ggplot2::labs(
        x     = "Degrees of Freedom (df)",
        y     = criterion,
        title = paste0("Knot Selection: ", criterion, " by df  (",
                       method, " spline)")
      ) +
      ggplot2::theme_minimal(base_size = 12)
    print(p)
  }

  message("Best df = ", best_df, "  (", criterion, " = ",
          round(crit_vals[which.min(crit_vals)], 2), ")")

  invisible(list(
    best_df      = best_df,
    search_table = tbl,
    criterion    = criterion,
    plot         = p
  ))
}

#' Intraclass correlation coefficients for a multilevel nl_fit model
#'
#' @description
#' Extracts variance components from a multilevel \code{nl_fit} object and
#' computes intraclass correlation coefficients (ICCs) for each grouping
#' level plus the residual.
#'
#' The ICC for grouping factor \eqn{g} is defined as:
#' \deqn{ICC_g = \frac{\sigma^2_g}{\sum_j \sigma^2_j + \sigma^2_\epsilon}}
#'
#' @param object An \code{nl_fit} object returned by \code{\link{nl_fit}}
#'   that was fitted with one or more \code{cluster} variables (i.e., a
#'   multilevel model fitted via \code{lme4::lmer()} or
#'   \code{lme4::glmer()}).
#' @param include_residual Logical; if \code{TRUE} (default), includes the
#'   residual variance component \code{ICC_resid} in the output so that all
#'   values sum to 1.
#'
#' @return
#' A named numeric vector of ICCs, one per grouping factor (named
#' \code{ICC_<groupname>}) plus \code{Residual} for the residual variance
#' (when \code{include_residual = TRUE}). All values sum to 1.
#'
#' @seealso \code{\link{nl_fit}}
#'
#' @examples
#' \dontrun{
#' fit <- nl_fit(
#'   data    = mydata,
#'   y       = "math_score",
#'   x       = "SES",
#'   cluster = c("id", "schid"),
#'   df      = 4
#' )
#' nl_icc(fit)
#' }
#'
#' @export
nl_icc <- function(object, include_residual = TRUE) {

  # ---------------------------------------------------------------------------
  # Input checks
  # ---------------------------------------------------------------------------
  if (!inherits(object, "nl_fit")) {
    stop(
      "`object` must be an 'nl_fit' object produced by nl_fit().",
      call. = FALSE
    )
  }

  mod <- object$model

  if (!inherits(mod, c("lmerMod", "glmerMod"))) {
    stop(
      "`nl_icc()` currently supports lme4 mixed models (lmerMod / glmerMod).",
      call. = FALSE
    )
  }

  # ---------------------------------------------------------------------------
  # Extract random-intercept variances
  # ---------------------------------------------------------------------------
  vc   <- lme4::VarCorr(mod)
  vcdf <- as.data.frame(vc)

  # Keep only random-intercept rows (var1 == "(Intercept)", var2 == NA)
  vcdf <- vcdf[
    !is.na(vcdf$grp) &
      vcdf$var1 == "(Intercept)" &
      is.na(vcdf$var2) &
      !is.na(vcdf$vcov),
    ,
    drop = FALSE
  ]

  if (!nrow(vcdf)) {
    stop("No random intercept variances found in the model.", call. = FALSE)
  }

  re_var <- stats::setNames(
    as.numeric(vcdf$vcov),
    as.character(vcdf$grp)
  )
  re_var <- re_var[!is.na(re_var)]

  # ---------------------------------------------------------------------------
  # Residual variance
  # ---------------------------------------------------------------------------
  resid_var <- NA_real_

  if (inherits(mod, "lmerMod")) {
    sc <- attr(vc, "sc")
    if (is.finite(sc)) resid_var <- sc^2
  } else {
    fam <- stats::family(mod)
    if (identical(fam$family, "binomial")) {
      resid_var <- if (identical(fam$link, "logit")) {
        (pi^2) / 3
      } else if (identical(fam$link, "probit")) {
        1
      } else {
        NA_real_
      }
    }
  }

  # ---------------------------------------------------------------------------
  # Compute ICCs
  # ---------------------------------------------------------------------------
  total <- sum(re_var, na.rm = TRUE)
  if (include_residual && is.finite(resid_var)) {
    total <- total + resid_var
  }

  if (!is.finite(total) || total <= 0) {
    stop(
      "Total variance is not finite or positive; cannot compute ICC.",
      call. = FALSE
    )
  }

  icc <- re_var / total

  if (include_residual && is.finite(resid_var)) {
    icc <- c(icc, Residual = resid_var / total)
  }

  icc
}

#' Compute first and second derivatives of the fitted spline curve
#'
#' @description
#' Uses numerical differentiation on the prediction grid to compute the
#' first derivative (slope / marginal effect), the second derivative
#' (curvature), and propagated confidence bands for both, using the
#' delta method from the \code{se.fit} column of \code{nl_predict()}.
#' Results are passed to \code{\link{nl_turning_points}} to identify
#' local turning points and inflection regions.
#'
#' @param pred_df A data frame returned by \code{\link{nl_predict}}.
#' @param x Character; name of the focal predictor column.
#' @param time Optional character; name of the time column. If present,
#'   derivatives are computed separately within each time level.
#' @param h Numeric; step size for finite-difference approximation.
#'   Default \code{NULL} uses 0.1 percent of the x range.
#' @param level Confidence level for derivative CIs. Default \code{0.95}.
#'
#' @return A data frame with columns \code{x} (the focal predictor),
#'   \code{time} (if applicable), \code{fit} (predicted outcome),
#'   \code{d1} (first derivative), \code{d1_lwr} and \code{d1_upr}
#'   (lower and upper CI for the first derivative), \code{d2} (second
#'   derivative), and \code{d2_lwr} and \code{d2_upr} (CI for the second
#'   derivative).
#'
#' @seealso \code{\link{nl_turning_points}}, \code{\link{nl_plot}}
#'
#' @export
nl_derivatives <- function(
    pred_df,
    x,
    time  = NULL,
    h     = NULL,
    level = 0.95
) {

  if (!is.data.frame(pred_df))
    stop("`pred_df` must be a data frame from nl_predict().", call. = FALSE)
  if (!x %in% names(pred_df))
    stop("`x` ('", x, "') not found in `pred_df`.", call. = FALSE)
  if (!"fit" %in% names(pred_df))
    stop("`pred_df` must contain a 'fit' column.", call. = FALSE)

  has_se <- "se.fit" %in% names(pred_df) && any(!is.na(pred_df$se.fit))
  crit   <- stats::qnorm(1 - (1 - level) / 2)

  # Auto-detect time column
  if (is.null(time)) {
    candidates <- c("TimePoint", "time", "Time", "wave", "Wave",
                    "occasion", "Occasion")
    hit <- candidates[candidates %in% names(pred_df)]
    if (length(hit)) time <- hit[1L]
  }

  .deriv_group <- function(df_sub) {
    xv  <- df_sub[[x]]
    yv  <- df_sub[["fit"]]
    sev <- if (has_se) df_sub[["se.fit"]] else rep(0, nrow(df_sub))
    n   <- length(xv)

    # Ensure sorted by x
    ord <- order(xv)
    xv  <- xv[ord]; yv <- yv[ord]; sev <- sev[ord]

    if (is.null(h)) h <- diff(range(xv, na.rm = TRUE)) * 0.001
    if (!is.finite(h) || h <= 0) h <- 1e-4

    # --- First derivative: central differences ---
    d1 <- rep(NA_real_, n)

    if (n >= 3L) {
      for (i in seq(2L, n - 1L)) {
        dx <- xv[i + 1L] - xv[i - 1L]
        if (is.finite(dx) && dx > 0)
          d1[i] <- (yv[i + 1L] - yv[i - 1L]) / dx
      }
      dx_f <- xv[2L] - xv[1L]
      dx_b <- xv[n] - xv[n - 1L]
      if (is.finite(dx_f) && dx_f > 0) d1[1L] <- (yv[2L] - yv[1L]) / dx_f
      if (is.finite(dx_b) && dx_b > 0) d1[n]  <- (yv[n] - yv[n - 1L]) / dx_b
    } else if (n == 2L) {
      dx <- xv[2L] - xv[1L]
      if (is.finite(dx) && dx > 0) d1[] <- (yv[2L] - yv[1L]) / dx
    }

    # --- Second derivative: finite difference of d1 ---
    d2 <- rep(NA_real_, n)

    if (n >= 3L) {
      for (i in seq(2L, n - 1L)) {
        dx <- xv[i + 1L] - xv[i - 1L]
        if (is.finite(dx) && dx > 0)
          d2[i] <- (d1[i + 1L] - d1[i - 1L]) / dx
      }
      dx_f <- xv[2L] - xv[1L]
      dx_b <- xv[n] - xv[n - 1L]
      if (is.finite(dx_f) && dx_f > 0)
        d2[1L] <- (d1[2L] - d1[1L]) / dx_f
      if (is.finite(dx_b) && dx_b > 0)
        d2[n]  <- (d1[n] - d1[n - 1L]) / dx_b
    }

    # --- Delta-method SE for d1 ---
    # Var(dy/dx) ~ (se_{i+1}^2 + se_{i-1}^2) / dx^2
    d1_se <- rep(NA_real_, n)

    if (has_se && n >= 3L) {
      for (i in seq(2L, n - 1L)) {
        dx <- xv[i + 1L] - xv[i - 1L]
        if (is.finite(dx) && dx > 0)
          d1_se[i] <- sqrt(sev[i + 1L]^2 + sev[i - 1L]^2) / abs(dx)
      }
      dx_f <- xv[2L] - xv[1L]
      dx_b <- xv[n] - xv[n - 1L]
      if (is.finite(dx_f) && dx_f > 0)
        d1_se[1L] <- sqrt(sev[2L]^2 + sev[1L]^2) / abs(dx_f)
      if (is.finite(dx_b) && dx_b > 0)
        d1_se[n]  <- sqrt(sev[n]^2 + sev[n - 1L]^2) / abs(dx_b)
    }

    # --- Delta-method SE for d2 (propagated from d1_se) ---
    d2_se <- rep(NA_real_, n)

    if (has_se && n >= 3L) {
      for (i in seq(2L, n - 1L)) {
        dx <- xv[i + 1L] - xv[i - 1L]
        if (is.finite(dx) && dx > 0 &&
            is.finite(d1_se[i + 1L]) && is.finite(d1_se[i - 1L]))
          d2_se[i] <- sqrt(d1_se[i + 1L]^2 + d1_se[i - 1L]^2) / abs(dx)
      }
    }

    data.frame(
      x_val  = xv,
      fit    = yv,
      d1     = d1,
      d1_lwr = d1 - crit * d1_se,
      d1_upr = d1 + crit * d1_se,
      d2     = d2,
      d2_lwr = d2 - crit * d2_se,
      d2_upr = d2 + crit * d2_se,
      stringsAsFactors = FALSE
    )
  }

  # Run within each time group if applicable
  if (!is.null(time) && time %in% names(pred_df)) {
    groups <- split(pred_df, pred_df[[time]])
    parts  <- lapply(names(groups), function(lv) {
      res         <- .deriv_group(groups[[lv]])
      res[[time]] <- lv
      res
    })
    out <- do.call(rbind, parts)
  } else {
    out <- .deriv_group(pred_df)
  }

  # Rename x_val back to the original predictor name
  names(out)[names(out) == "x_val"] <- x

  if (requireNamespace("dplyr", quietly = TRUE)) out <- dplyr::as_tibble(out)
  out
}


#' Identify turning points and inflection regions
#'
#' @description
#' From the derivative table returned by \code{\link{nl_derivatives}}, finds
#' turning points (values of \code{x} where the first derivative crosses zero),
#' inflection regions (intervals where the second derivative changes sign),
#' and slope regions (contiguous stretches where the curve is increasing,
#' decreasing, or flat).
#'
#' @param deriv_df A data frame returned by \code{\link{nl_derivatives}}.
#' @param x Character; name of the focal predictor column.
#' @param time Optional character; name of the time column. Turning points
#'   are identified separately within each time level when supplied.
#' @param tol Numeric tolerance for near-zero detection of d1. Default
#'   \code{1e-6}.
#'
#' @return A named list with three elements.
#'   \code{turning_points} is a data frame with columns \code{x},
#'   \code{time} (if applicable), \code{fit}, and \code{type}
#'   (one of \code{"maximum"}, \code{"minimum"}, or \code{"saddle"}).
#'   \code{inflection_regions} is a data frame with columns
#'   \code{x_start}, \code{x_end}, \code{time}, and \code{direction}
#'   (\code{"concave_to_convex"} or \code{"convex_to_concave"}).
#'   \code{slope_regions} is a data frame with columns \code{x_start},
#'   \code{x_end}, \code{direction}, and \code{time}.
#'
#' @seealso \code{\link{nl_derivatives}}, \code{\link{nl_plot}}
#'
#' @export
nl_turning_points <- function(
    deriv_df,
    x,
    time = NULL,
    tol  = 1e-6
) {

  if (!is.data.frame(deriv_df))
    stop("`deriv_df` must be a data frame from nl_derivatives().", call. = FALSE)
  if (!x %in% names(deriv_df))
    stop("`x` ('", x, "') not found in `deriv_df`.", call. = FALSE)

  # Auto-detect time column
  if (is.null(time)) {
    candidates <- c("TimePoint", "time", "Time", "wave", "Wave",
                    "occasion", "Occasion")
    hit <- candidates[candidates %in% names(deriv_df)]
    if (length(hit)) time <- hit[1L]
  }

  .find_tp <- function(df_sub, time_val = NULL) {

    xv  <- df_sub[[x]]
    d1  <- df_sub[["d1"]]
    d2  <- df_sub[["d2"]]
    fit <- df_sub[["fit"]]
    n   <- length(xv)

    # --- Turning points: d1 crosses zero ---
    tp_list <- list()

    if (n >= 2L) {
      for (i in seq_len(n - 1L)) {
        if (is.finite(d1[i]) && is.finite(d1[i + 1L])) {

          if (d1[i] * d1[i + 1L] < 0) {
            x_cross   <- xv[i] - d1[i] * (xv[i + 1L] - xv[i]) /
              (d1[i + 1L] - d1[i])
            fit_cross <- fit[i] + (fit[i + 1L] - fit[i]) *
              (x_cross - xv[i]) / (xv[i + 1L] - xv[i])
            d2_cross  <- if (is.finite(d2[i])) d2[i] else NA_real_
            tp_type   <- if (!is.na(d2_cross)) {
              if (d2_cross < 0) "maximum" else "minimum"
            } else {
              if (d1[i] > 0) "maximum" else "minimum"
            }
            tp_list[[length(tp_list) + 1L]] <- data.frame(
              x_cross = x_cross,
              fit     = fit_cross,
              type    = tp_type,
              stringsAsFactors = FALSE
            )

          } else if (abs(d1[i]) < tol) {
            tp_type <- if (is.finite(d2[i])) {
              if (d2[i] < 0) "maximum" else if (d2[i] > 0) "minimum" else "saddle"
            } else "saddle"
            tp_list[[length(tp_list) + 1L]] <- data.frame(
              x_cross = xv[i], fit = fit[i], type = tp_type,
              stringsAsFactors = FALSE
            )
          }
        }
      }
    }

    tp_df <- if (length(tp_list)) {
      res <- do.call(rbind, tp_list)
      names(res)[names(res) == "x_cross"] <- x
      if (!is.null(time_val)) res[[time]] <- time_val
      res
    } else {
      base <- data.frame(
        matrix(ncol = 3, nrow = 0,
               dimnames = list(NULL, c(x, "fit", "type")))
      )
      if (!is.null(time_val)) base[[time]] <- character(0)
      base
    }

    # --- Inflection regions: d2 changes sign ---
    infl_list <- list()

    if (n >= 2L && "d2" %in% names(df_sub)) {
      for (i in seq_len(n - 1L)) {
        if (is.finite(d2[i]) && is.finite(d2[i + 1L]) &&
            d2[i] * d2[i + 1L] < 0) {
          dir <- if (d2[i] < 0) "concave_to_convex" else "convex_to_concave"
          infl_list[[length(infl_list) + 1L]] <- data.frame(
            x_start   = xv[i],
            x_end     = xv[i + 1L],
            direction = dir,
            stringsAsFactors = FALSE
          )
        }
      }
    }

    infl_df <- if (length(infl_list)) {
      res <- do.call(rbind, infl_list)
      if (!is.null(time_val)) res[[time]] <- time_val
      res
    } else {
      base <- data.frame(
        x_start   = numeric(0),
        x_end     = numeric(0),
        direction = character(0),
        stringsAsFactors = FALSE
      )
      if (!is.null(time_val)) base[[time]] <- character(0)
      base
    }

    # --- Slope regions: contiguous increasing / decreasing / flat ---
    slope_dir <- ifelse(
      is.na(d1), "flat",
      ifelse(d1 > tol, "increasing",
             ifelse(d1 < -tol, "decreasing", "flat"))
    )
    rle_res  <- rle(slope_dir)
    ends     <- cumsum(rle_res$lengths)
    starts   <- c(1L, ends[-length(ends)] + 1L)

    slope_df <- data.frame(
      x_start   = xv[starts],
      x_end     = xv[ends],
      direction = rle_res$values,
      stringsAsFactors = FALSE
    )
    if (!is.null(time_val)) slope_df[[time]] <- time_val

    list(tp = tp_df, infl = infl_df, slope = slope_df)
  }

  # Run within each time group if applicable
  if (!is.null(time) && time %in% names(deriv_df)) {
    groups    <- split(deriv_df, deriv_df[[time]])
    res_list  <- lapply(names(groups), function(lv) .find_tp(groups[[lv]], lv))
    tp_all    <- do.call(rbind, lapply(res_list, `[[`, "tp"))
    infl_all  <- do.call(rbind, lapply(res_list, `[[`, "infl"))
    slope_all <- do.call(rbind, lapply(res_list, `[[`, "slope"))
  } else {
    res       <- .find_tp(deriv_df)
    tp_all    <- res$tp
    infl_all  <- res$infl
    slope_all <- res$slope
  }

  list(
    turning_points     = tp_all,
    inflection_regions = infl_all,
    slope_regions      = slope_all
  )
}

#' Print method for nl_fit objects
#'
#' @description
#' Compact console display for objects returned by \code{\link{nl_fit}}.
#' Shows key metadata (method, outcome, predictor, time, clustering, family,
#' and controls) and the fitted model formula.
#'
#' @param x An object of class \code{nl_fit}.
#' @param ... Further arguments (currently ignored).
#'
#' @return \code{x} invisibly.
#'
#' @method print nl_fit
#' @export
print.nl_fit <- function(x, ...) {

  if (!inherits(x, "nl_fit")) {
    stop("`x` must be an object of class 'nl_fit'.", call. = FALSE)
  }

  cat("MultiSpline fit  (nl_fit object)\n")
  cat("----------------------------------\n")
  cat("Method:  ", x$method, "\n", sep = "")
  cat("Outcome: ", x$y,      "\n", sep = "")
  cat("Focal x: ", x$x,      "\n", sep = "")

  if (!is.null(x$time)) {
    cat("Time:    ", x$time, "\n", sep = "")
  } else {
    cat("Time:     (none)\n")
  }

  if (is.null(x$cluster) || length(x$cluster) == 0L) {
    cat("Cluster:  (none; single-level)\n")
  } else {
    cat("Cluster: ", paste(x$cluster, collapse = " > "), "\n", sep = "")
  }

  if (!is.null(x$family) && !is.null(x$family$family)) {
    cat("Family:  ", x$family$family, "\n", sep = "")
  }

  if (!is.null(x$controls) && length(x$controls) > 0L) {
    cat("Controls:", paste(x$controls, collapse = ", "), "\n")
  }

  if (!is.null(x$call)) {
    cat("\nCall:\n")
    print(x$call)
  }

  cat("\nModel formula:\n")
  if (!is.null(x$formula)) {
    print(x$formula)
  } else if (!is.null(x$model)) {
    print(stats::formula(x$model))
  } else {
    cat("(formula unavailable)\n")
  }

  cat("\nNext steps (v2):\n")
  cat("  nl_summary(fit)               # tidy coefficient table\n")
  cat("  nl_predict(fit)               # prediction grid with CI\n")
  cat("  nl_plot(pred_df, x = ...)     # trajectory / slope / curvature\n")
  cat("  nl_derivatives(pred_df, x)   # d1 / d2 with confidence bands\n")
  cat("  nl_turning_points(deriv_df)   # maxima, minima, inflection regions\n")
  cat("  nl_compare(fit)               # spline vs linear vs polynomial\n")
  cat("  nl_r2(fit)                    # R2 / multilevel variance decomposition\n")
  cat("  nl_knots(data, ...)           # explore df / knot selection\n")

  if (!is.null(x$cluster) && length(x$cluster) > 0L) {
    cat("  nl_icc(fit)                   # intraclass correlation\n")
    cat("  nl_het(fit)                   # cluster heterogeneity in nonlinear effect\n")
  }

  if (!is.null(x$df_search)) {
    cat(sprintf("\n  Auto-selected df = %d  (best %s = %.2f)\n",
                x$df_selected,
                names(x$df_search)[2],
                min(x$df_search[[2]], na.rm = TRUE)))
  }

  invisible(x)
}


#' Summary method for nl_fit objects
#'
#' @description
#' Produces a tidy coefficient table via \code{\link{nl_summary}} and wraps
#' it in a \code{summary_nl_fit} object for pretty printing.
#'
#' @param object An \code{nl_fit} object returned by \code{\link{nl_fit}}.
#' @param digits Number of decimal places for rounding. Default \code{3}.
#' @param pvals Logical; if \code{TRUE}, attempts to include p-values.
#'   Default \code{TRUE}.
#' @param df_method For \code{lmerMod}: \code{"satterthwaite"} (requires
#'   \pkg{lmerTest}) or \code{"none"}. Default \code{"satterthwaite"}.
#' @param ... Further arguments passed to \code{\link{nl_summary}}.
#'
#' @return
#' An object of class \code{summary_nl_fit} containing \code{call},
#' \code{formula}, \code{method}, and \code{table}.
#'
#' @method summary nl_fit
#' @export
summary.nl_fit <- function(
    object,
    digits    = 3,
    pvals     = TRUE,
    df_method = c("satterthwaite", "none"),
    ...
) {

  if (!inherits(object, "nl_fit")) {
    stop("`object` must be an object of class 'nl_fit'.", call. = FALSE)
  }

  df_method <- match.arg(df_method)

  tab <- nl_summary(
    object,
    digits    = digits,
    pvals     = pvals,
    df_method = df_method,
    ...
  )

  out <- list(
    call    = object$call %||% NULL,
    formula = tryCatch(
      stats::formula(object$model),
      error = function(e) NULL
    ),
    method  = object$method,
    table   = tab
  )

  class(out) <- "summary_nl_fit"
  out
}


#' Print method for summary_nl_fit objects
#'
#' @description
#' Prints a \code{summary_nl_fit} object returned by
#' \code{\link{summary.nl_fit}}.
#'
#' @param x A \code{summary_nl_fit} object.
#' @param ... Further arguments passed to \code{print}.
#'
#' @return \code{x} invisibly.
#'
#' @method print summary_nl_fit
#' @export
print.summary_nl_fit <- function(x, ...) {

  if (!inherits(x, "summary_nl_fit")) {
    stop("`x` must be an object of class 'summary_nl_fit'.", call. = FALSE)
  }

  cat("Summary of nl_fit object\n")
  cat("  Method:", x$method, "\n")

  if (!is.null(x$formula)) {
    cat("\nFormula:\n")
    print(x$formula)
  }

  cat("\nCoefficient table:\n")
  if (!is.null(x$table)) {
    print(as.data.frame(x$table), row.names = FALSE, ...)
  } else {
    cat("(table unavailable)\n")
  }

  invisible(x)
}

#' Cluster heterogeneity in nonlinear effects
#'
#' @description
#' Quantifies and visualises how much the nonlinear relationship between
#' \code{x} and \code{y} varies across clusters, using a model with random
#' spline slopes.
#'
#' The function:
#' \enumerate{
#'   \item Refits the model with random spline slopes (if not already fitted
#'     with \code{random_slope = TRUE}).
#'   \item Extracts cluster-specific predicted curves (BLUPs).
#'   \item Returns a heterogeneity summary: variance of slopes across clusters
#'     at each x value, and a plot of the cluster-specific trajectories.
#'   \item Performs a likelihood-ratio test comparing the random-slope model
#'     against a random-intercept-only model to assess whether heterogeneity
#'     is statistically significant.
#' }
#'
#' @param object An \code{nl_fit} object fitted with a single \code{cluster}
#'   variable (and either \code{random_slope = TRUE} or \code{FALSE}; in the
#'   latter case the model is refit internally).
#' @param n_clusters_plot Maximum number of cluster curves to display in the
#'   trajectory plot. Default \code{30L}. Use \code{Inf} for all clusters.
#' @param x_seq Optional numeric vector of x values for prediction.
#' @param level Confidence level for the population-mean ribbon. Default
#'   \code{0.95}.
#' @param plot Logical; print the trajectory plot. Default \code{TRUE}.
#' @param seed Integer seed for reproducibility when sub-sampling clusters.
#'   Default \code{42L}.
#'
#' @return A list (invisibly) with:
#'   \describe{
#'     \item{\code{trajectory_df}}{Long data frame of cluster-specific
#'       predicted values.}
#'     \item{\code{slope_variance}}{Named numeric: SD of first-derivative
#'       estimates across clusters at each x value.}
#'     \item{\code{lrt}}{LRT comparing random-slope vs random-intercept model
#'       (\code{NULL} if \code{random_slope = TRUE} was supplied).}
#'     \item{\code{plot}}{A \code{ggplot} object.}
#'   }
#'
#' @seealso \code{\link{nl_fit}}, \code{\link{nl_derivatives}}
#'
#' @export
nl_het <- function(
    object,
    n_clusters_plot = 30L,
    x_seq           = NULL,
    level           = 0.95,
    plot            = TRUE,
    seed            = 42L
) {

  if (!inherits(object, "nl_fit"))
    stop("`object` must be an 'nl_fit' object.", call. = FALSE)
  if (is.null(object$cluster) || length(object$cluster) != 1L)
    stop("`nl_het()` currently requires exactly one cluster variable.", call. = FALSE)

  cluster_var <- object$cluster[1L]
  mod_ri      <- object$model   # random-intercept model (or the supplied one)

  # Retrieve original data from the model frame
  mf    <- tryCatch(stats::model.frame(mod_ri), error = function(e) NULL)
  if (is.null(mf))
    stop("Cannot extract model frame.", call. = FALSE)

  # Attempt to recover full data (with cluster column)
  orig_data <- tryCatch(
    eval(object$call$data, envir = parent.frame()),
    error = function(e) NULL
  )
  fit_data <- if (!is.null(orig_data) && cluster_var %in% names(orig_data)) {
    orig_data
  } else {
    # Try attaching cluster from model object
    tryCatch({
      re_df  <- as.data.frame(lme4::ranef(mod_ri))
      cl_ids <- rownames(lme4::ranef(mod_ri)[[cluster_var]])
      stop("Cannot recover original data with cluster column. ",
           "Please pass the data via `nl_fit(data = ...)` using a named object.",
           call. = FALSE)
    }, error = function(e) NULL)
  }
  if (is.null(fit_data))
    stop("Cannot recover original data.", call. = FALSE)

  # ---  Refit with random slopes if needed  ---
  lrt <- NULL
  if (!isTRUE(object$random_slope)) {
    message("Refitting with random spline slopes to assess heterogeneity ...")
    mod_rs <- tryCatch(
      .build_and_fit(
        data = fit_data, y = object$y, x = object$x,
        time = object$time, cluster = object$cluster,
        nested = FALSE, controls = object$controls,
        method = object$method, df = object$df,
        k = object$k, bs_degree = object$bs_degree,
        random_slope = TRUE, family = object$family, dots = list()
      ),
      error = function(e) {
        message("Random-slope model failed: ", e$message,
                "\nReturning heterogeneity based on random-intercept BLUPs.")
        NULL
      }
    )
    if (!is.null(mod_rs)) {
      lrt <- tryCatch(
        stats::anova(mod_ri, mod_rs),
        error = function(e) NULL
      )
      mod_rs_use <- mod_rs
    } else {
      mod_rs_use <- mod_ri
    }
  } else {
    mod_rs_use <- mod_ri
  }

  # ---  x grid  ---
  xi <- object$x_info
  if (is.null(x_seq)) {
    lo   <- xi$q01; hi <- xi$q99
    if (!is.finite(lo) || !is.finite(hi) || hi <= lo) { lo <- xi$min; hi <- xi$max }
    x_seq <- seq(lo, hi, length.out = 100L)
  }

  # ---  Cluster IDs  ---
  re         <- lme4::ranef(mod_rs_use)[[cluster_var]]
  all_ids    <- rownames(re)
  n_total    <- length(all_ids)
  set.seed(seed)
  plot_ids   <- if (is.finite(n_clusters_plot) && n_clusters_plot < n_total) {
    sample(all_ids, min(as.integer(n_clusters_plot), n_total))
  } else all_ids

  # ---  Cluster-specific predictions  ---
  x_var    <- object$x
  controls <- object$controls

  # Build newdata for each cluster
  traj_list <- lapply(plot_ids, function(id) {
    nd          <- data.frame(x_seq)
    names(nd)   <- x_var
    nd[[cluster_var]] <- id
    if (!is.null(controls)) {
      for (cn in controls)
        nd[[cn]] <- object$control_defaults[[cn]]
    }
    # Preserve factor levels for cluster variable
    cl_vals <- fit_data[[cluster_var]]
    if (is.factor(cl_vals))
      nd[[cluster_var]] <- factor(id, levels = levels(cl_vals))

    nd$fit <- tryCatch(
      as.numeric(stats::predict(mod_rs_use, newdata = nd, re.form = NULL,
                                allow.new.levels = TRUE)),
      error = function(e) rep(NA_real_, nrow(nd))
    )
    nd$cluster_id <- id
    nd[, c(x_var, "fit", "cluster_id")]
  })
  traj_df <- do.call(rbind, traj_list)

  # Population-level ribbon
  pop_pred <- nl_predict(object, x_seq = x_seq, se = TRUE, level = level,
                         re_form = NA)

  # Slope variance across clusters
  slope_sd <- tapply(traj_df$fit, traj_df[[x_var]], stats::sd, na.rm = TRUE)

  # ---  Plot  ---
  p <- ggplot2::ggplot() +
    ggplot2::geom_line(
      data    = traj_df,
      mapping = ggplot2::aes(
        x     = .data[[x_var]],
        y     = .data[["fit"]],
        group = .data[["cluster_id"]]
      ),
      alpha    = 0.25,
      linewidth = 0.35,
      color    = "steelblue"
    ) +
    ggplot2::geom_ribbon(
      data    = pop_pred,
      mapping = ggplot2::aes(
        x    = .data[[x_var]],
        ymin = .data[["lwr"]],
        ymax = .data[["upr"]]
      ),
      alpha = 0.30,
      fill  = "#D7191C"
    ) +
    ggplot2::geom_line(
      data    = pop_pred,
      mapping = ggplot2::aes(
        x = .data[[x_var]],
        y = .data[["fit"]]
      ),
      linewidth = 1.1,
      color    = "#D7191C"
    ) +
    ggplot2::labs(
      x     = x_var,
      y     = paste("Predicted", object$y),
      title = paste0("Cluster Heterogeneity in the Effect of ", x_var),
      subtitle = paste0(
        "Blue = cluster-specific curves  |  Red = population mean \u00b1 ",
        round(level * 100), "% CI\n",
        "Displaying ", length(plot_ids), " of ", n_total, " clusters"
      )
    ) +
    ggplot2::theme_minimal(base_size = 12) +
    ggplot2::theme(plot.subtitle = ggplot2::element_text(size = 9, color = "grey40"))

  if (plot) print(p)

  if (!is.null(lrt)) {
    cat("\n--- LRT: Random slopes vs Random intercepts ---\n")
    print(lrt)
  }

  invisible(list(
    trajectory_df  = traj_df,
    slope_variance = slope_sd,
    lrt            = lrt,
    plot           = p
  ))
}

#' Generate predictions from an nl_fit model
#'
#' @description
#' Creates a prediction data frame over a grid of the focal predictor \code{x}
#' (and optionally over \code{time}), holding control variables at typical
#' values. For mixed models, predictions default to population-level curves
#' (random effects set to zero).
#'
#' v2 improvements:
#' \itemize{
#'   \item \strong{CI for glmerMod}: approximate confidence intervals are
#'     computed via the parametric bootstrap or the delta method. Set
#'     \code{glmer_ci = "delta"} (default, fast) or \code{"boot"} (more
#'     accurate, slower).
#'   \item \strong{Cluster-specific predictions}: set \code{re_form = NULL}
#'     to include random effects in the predictions.
#' }
#'
#' @param object An \code{nl_fit} object.
#' @param x_seq Optional numeric vector of x values. If \code{NULL}, 200
#'   evenly-spaced points between the 1st and 99th percentiles.
#' @param time_levels Optional vector of time levels.
#' @param controls_fixed Optional named list of fixed control values.
#' @param se Logical; include SEs and CIs. Default \code{TRUE}.
#' @param level Confidence level. Default \code{0.95}.
#' @param re_form For mixed models: \code{NA} (population-level, default) or
#'   \code{NULL} (include random effects).
#' @param glmer_ci Method for glmerMod CIs: \code{"delta"} (default) or
#'   \code{"boot"} (parametric bootstrap).
#' @param n_boot Number of bootstrap replicates when \code{glmer_ci = "boot"}.
#'   Default \code{500}.
#' @param ... Reserved for future use.
#'
#' @return A data frame with columns for the focal predictor, time (if any),
#'   controls at fixed values, \code{fit}, \code{se.fit}, \code{lwr},
#'   and \code{upr}.
#'
#' @seealso \code{\link{nl_fit}}, \code{\link{nl_plot}},
#'   \code{\link{nl_derivatives}}
#'
#' @export
nl_predict <- function(
    object,
    x_seq       = NULL,
    time_levels = NULL,
    controls_fixed = NULL,
    se          = TRUE,
    level       = 0.95,
    re_form     = NA,
    glmer_ci    = c("delta", "boot"),
    n_boot      = 500L,
    ...
) {

  if (!inherits(object, "nl_fit"))
    stop("`object` must be of class 'nl_fit'.", call. = FALSE)
  if (!is.numeric(level) || level <= 0 || level >= 1)
    stop("`level` must be between 0 and 1.", call. = FALSE)

  glmer_ci <- match.arg(glmer_ci)
  mod      <- object$model
  x        <- object$x
  time_var <- object$time
  controls <- object$controls

  # ---------------------------------------------------------------------------
  # 1) Build x grid
  # ---------------------------------------------------------------------------
  xi <- object$x_info
  if (is.null(x_seq)) {
    lo <- xi$q01; hi <- xi$q99
    if (!is.finite(lo) || !is.finite(hi) || hi <= lo) { lo <- xi$min; hi <- xi$max }
    if (!is.finite(lo) || !is.finite(hi) || hi <= lo)
      stop("Cannot construct x grid: x range is not finite.", call. = FALSE)
    x_seq <- seq(lo, hi, length.out = 200L)
  } else {
    if (!is.numeric(x_seq) || length(x_seq) < 2L)
      stop("`x_seq` must be a numeric vector of length >= 2.", call. = FALSE)
    x_seq <- as.numeric(x_seq)
  }

  # ---------------------------------------------------------------------------
  # 2) Time levels
  # ---------------------------------------------------------------------------
  levels_info <- object$levels_info
  if (!is.null(time_var) && is.null(time_levels)) {
    if (!is.null(levels_info[[time_var]])) {
      time_levels <- levels_info[[time_var]]
    } else {
      mf <- try(stats::model.frame(mod), silent = TRUE)
      if (!inherits(mf, "try-error") && time_var %in% names(mf))
        time_levels <- unique(mf[[time_var]])
      else
        stop("Cannot determine `time_levels`. Please supply them.", call. = FALSE)
    }
  }

  # ---------------------------------------------------------------------------
  # 3) Build newdata
  # ---------------------------------------------------------------------------
  if (is.null(time_levels)) {
    newdata        <- data.frame(x_seq, stringsAsFactors = FALSE)
    names(newdata) <- x
  } else {
    grid           <- expand.grid(x_seq = x_seq, tl = time_levels,
                                  KEEP.OUT.ATTRS = FALSE, stringsAsFactors = FALSE)
    names(grid)    <- c(x, time_var)
    newdata        <- grid
    if (!is.null(levels_info[[time_var]]))
      newdata[[time_var]] <- factor(newdata[[time_var]], levels = levels_info[[time_var]])
  }

  # Fill controls
  if (!is.null(controls)) {
    for (cn in controls) {
      val <- if (!is.null(controls_fixed[[cn]])) controls_fixed[[cn]] else
             object$control_defaults[[cn]]
      if (is.null(val)) stop("Cannot determine default for control `", cn, "`.", call. = FALSE)
      if (!is.null(levels_info[[cn]]))
        newdata[[cn]] <- factor(val, levels = levels_info[[cn]])
      else
        newdata[[cn]] <- val
    }
  }

  # ---------------------------------------------------------------------------
  # 4) Predict by model type
  # ---------------------------------------------------------------------------
  alpha <- 1 - level
  crit  <- stats::qnorm(1 - alpha / 2)

  fit_vals <- se_vals <- lwr <- upr <- NULL

  if (inherits(mod, "lm")) {
    # ------- OLS / lm -------
    pred     <- stats::predict(mod, newdata = newdata, se.fit = se)
    if (isTRUE(se)) {
      fit_vals <- as.numeric(pred$fit)
      se_vals  <- as.numeric(pred$se.fit)
      lwr      <- fit_vals - crit * se_vals
      upr      <- fit_vals + crit * se_vals
    } else {
      fit_vals <- as.numeric(pred)
      se_vals  <- lwr <- upr <- rep(NA_real_, length(fit_vals))
    }

  } else if (inherits(mod, "gam")) {
    # ------- GAM -------
    pred <- mgcv::predict.gam(mod, newdata = newdata, se.fit = se,
                              type = "response")
    if (isTRUE(se)) {
      fit_vals <- as.numeric(pred$fit)
      se_vals  <- as.numeric(pred$se.fit)
      lwr      <- fit_vals - crit * se_vals
      upr      <- fit_vals + crit * se_vals
    } else {
      fit_vals <- as.numeric(pred)
      se_vals  <- lwr <- upr <- rep(NA_real_, length(fit_vals))
    }

  } else if (inherits(mod, "lmerMod")) {
    # ------- LMM (lmer) -------
    fit_vals <- as.numeric(stats::predict(mod, newdata = newdata, re.form = re_form))
    if (isTRUE(se)) {
      ff_nobars <- tryCatch(
        {
          if (requireNamespace("reformulas", quietly = TRUE))
            reformulas::nobars(stats::formula(mod))
          else
            lme4::nobars(stats::formula(mod))
        },
        error = function(e) NULL
      )
      if (!is.null(ff_nobars)) {
        tt <- stats::delete.response(stats::terms(ff_nobars))
        X  <- tryCatch(stats::model.matrix(tt, newdata), error = function(e) NULL)
        V  <- as.matrix(stats::vcov(mod))
        if (!is.null(X)) {
          common  <- intersect(colnames(X), colnames(V))
          if (length(common) > 0L) {
            Xc      <- X[, common, drop = FALSE]
            Vc      <- V[common, common, drop = FALSE]
            se_vals <- sqrt(pmax(rowSums((Xc %*% Vc) * Xc), 0))
            lwr     <- fit_vals - crit * se_vals
            upr     <- fit_vals + crit * se_vals
          } else {
            se_vals <- lwr <- upr <- rep(NA_real_, length(fit_vals))
          }
        } else {
          se_vals <- lwr <- upr <- rep(NA_real_, length(fit_vals))
        }
      } else {
        se_vals <- lwr <- upr <- rep(NA_real_, length(fit_vals))
      }
    } else {
      se_vals <- lwr <- upr <- rep(NA_real_, length(fit_vals))
    }

  } else if (inherits(mod, "glmerMod")) {
    # ------- GLMM (glmer) — v2: CI via delta method or bootstrap -------
    fit_vals <- as.numeric(
      stats::predict(mod, newdata = newdata, re.form = re_form, type = "response")
    )
    if (isTRUE(se)) {
      if (glmer_ci == "delta") {
        # Delta method on link scale → back-transform
        fit_link <- as.numeric(
          stats::predict(mod, newdata = newdata, re.form = re_form, type = "link")
        )
        ff <- stats::formula(mod)
        ff_nobars <- tryCatch(
          if (requireNamespace("reformulas", quietly = TRUE))
            reformulas::nobars(ff) else lme4::nobars(ff),
          error = function(e) NULL
        )
        if (!is.null(ff_nobars)) {
          tt     <- stats::delete.response(stats::terms(ff_nobars))
          X      <- tryCatch(stats::model.matrix(tt, newdata), error = function(e) NULL)
          V      <- as.matrix(stats::vcov(mod))
          if (!is.null(X)) {
            common <- intersect(colnames(X), colnames(V))
            if (length(common) > 0L) {
              Xc       <- X[, common, drop = FALSE]
              Vc       <- V[common, common, drop = FALSE]
              se_link  <- sqrt(pmax(rowSums((Xc %*% Vc) * Xc), 0))
              inv_link <- mod@resp$family$linkinv
              lwr      <- inv_link(fit_link - crit * se_link)
              upr      <- inv_link(fit_link + crit * se_link)
              se_vals  <- (upr - lwr) / (2 * crit)
            } else {
              se_vals <- lwr <- upr <- rep(NA_real_, length(fit_vals))
            }
          } else {
            se_vals <- lwr <- upr <- rep(NA_real_, length(fit_vals))
          }
        } else {
          se_vals <- lwr <- upr <- rep(NA_real_, length(fit_vals))
        }

      } else {
        # Parametric bootstrap CIs
        message("Computing bootstrap CIs for glmerMod (n_boot = ", n_boot, ")...")
        boot_mat <- tryCatch({
          lme4::bootMer(
            mod,
            FUN = function(m) as.numeric(
              stats::predict(m, newdata = newdata, re.form = re_form,
                             type = "response")
            ),
            nsim      = n_boot,
            use.u     = FALSE,
            type      = "parametric",
            .progress = "none"
          )$t
        }, error = function(e) {
          warning("Bootstrap failed: ", conditionMessage(e), call. = FALSE)
          NULL
        })
        if (!is.null(boot_mat)) {
          lwr     <- apply(boot_mat, 2L, stats::quantile,
                           probs = alpha / 2,     na.rm = TRUE)
          upr     <- apply(boot_mat, 2L, stats::quantile,
                           probs = 1 - alpha / 2, na.rm = TRUE)
          se_vals <- apply(boot_mat, 2L, stats::sd, na.rm = TRUE)
        } else {
          se_vals <- lwr <- upr <- rep(NA_real_, length(fit_vals))
        }
      }
    } else {
      se_vals <- lwr <- upr <- rep(NA_real_, length(fit_vals))
    }

  } else {
    stop("Unknown underlying model class inside `nl_fit`.", call. = FALSE)
  }

  # ---------------------------------------------------------------------------
  # 5) Assemble output
  # ---------------------------------------------------------------------------
  out        <- newdata
  out$fit    <- fit_vals
  out$se.fit <- se_vals
  out$lwr    <- lwr
  out$upr    <- upr

  if (requireNamespace("dplyr", quietly = TRUE)) out <- dplyr::as_tibble(out)
  out
}

#' Tidy coefficient table for an nl_fit model
#'
#' @description
#' Produces a tidy coefficient table for a model fitted by
#' \code{\link{nl_fit}}. For linear mixed models (\code{lmerMod}), optional
#' p-values and denominator degrees of freedom are obtained via
#' \pkg{lmerTest} (Satterthwaite method). For GLMMs (\code{glmerMod}),
#' z-tests are reported from \code{summary()}.
#'
#' @param object An \code{nl_fit} object returned by \code{\link{nl_fit}}.
#' @param digits Integer; number of decimal places for rounding. If
#'   \code{NULL}, no rounding is applied. Default \code{3}.
#' @param pvals Logical; if \code{TRUE}, attempts to include p-values.
#'   Default \code{TRUE}.
#' @param df_method For \code{lmerMod} with \code{pvals = TRUE}:
#'   \code{"satterthwaite"} (requires \pkg{lmerTest}) or \code{"none"}.
#'   Default \code{"satterthwaite"}.
#'
#' @return
#' A data frame (or tibble) with columns: \code{Term}, \code{Estimate},
#' \code{Std.Error}, \code{df} (if available), \code{statistic}, and
#' \code{p.value} (if available).
#'
#' @seealso \code{\link{nl_fit}}, \code{\link{summary.nl_fit}}
#'
#' @export
nl_summary <- function(
    object,
    digits    = 3,
    pvals     = TRUE,
    df_method = c("satterthwaite", "none")
) {

  # ---------------------------------------------------------------------------
  # Input checks
  # ---------------------------------------------------------------------------
  if (!inherits(object, "nl_fit")) {
    stop(
      "`object` must be an 'nl_fit' object produced by nl_fit().",
      call. = FALSE
    )
  }

  df_method <- match.arg(df_method)
  mod       <- object$model

  maybe_round <- function(x) {
    if (is.null(digits) || !is.numeric(x)) return(x)
    round(x, digits)
  }

  out <- NULL

  # ---------------------------------------------------------------------------
  # IMPORTANT: "gam" must be checked BEFORE "lm" because mgcv::gam objects
  # inherit from c("gam", "glm", "lm") — checking "lm" first would match GAM
  # models incorrectly and use s$coefficients instead of s$p.table/s$s.table.
  # ---------------------------------------------------------------------------

  # ---------------------------------------------------------------------------
  # 1) gam  <-- MUST come before lm
  # ---------------------------------------------------------------------------
  if (inherits(mod, "gam")) {

    s <- summary(mod)

    # -- Parametric terms (p.table) --
    # Guard against NULL or zero-row table (e.g. smooth-only models)
    p_out <- NULL
    if (!is.null(s$p.table) && nrow(s$p.table) > 0L) {
      ptab  <- as.data.frame(s$p.table)
      t_col <- grep("t value|z value", names(ptab), value = TRUE)[1L]
      p_col <- if ("Pr(>|t|)" %in% names(ptab)) {
        "Pr(>|t|)"
      } else if ("Pr(>|z|)" %in% names(ptab)) {
        "Pr(>|z|)"
      } else {
        NA_character_
      }

      p_out <- data.frame(
        Term      = rownames(ptab),
        Estimate  = ptab[["Estimate"]],
        Std.Error = ptab[["Std. Error"]],
        df        = NA_real_,
        statistic = if (!is.na(t_col)) ptab[[t_col]] else NA_real_,
        p.value   = if (!is.na(p_col)) ptab[[p_col]] else NA_real_,
        row.names = NULL,
        stringsAsFactors = FALSE
      )
    }

    # -- Smooth terms (s.table) --
    # Guard against NULL or zero-row table
    s_out <- NULL
    if (!is.null(s$s.table) && nrow(s$s.table) > 0L) {
      stab     <- as.data.frame(s$s.table)
      stat_col <- if ("F" %in% names(stab)) {
        "F"
      } else if ("Chi.sq" %in% names(stab)) {
        "Chi.sq"
      } else {
        names(stab)[1L]
      }
      p_col <- if ("p-value" %in% names(stab)) {
        "p-value"
      } else if ("p.value" %in% names(stab)) {
        "p.value"
      } else {
        NA_character_
      }

      s_out <- data.frame(
        Term      = rownames(stab),
        Estimate  = NA_real_,
        Std.Error = NA_real_,
        df        = if ("Ref.df" %in% names(stab)) stab[["Ref.df"]] else
          NA_real_,
        statistic = stab[[stat_col]],
        p.value   = if (!is.na(p_col)) stab[[p_col]] else NA_real_,
        row.names = NULL,
        stringsAsFactors = FALSE
      )
    }

    # Safely combine: either or both sections may be NULL
    parts <- Filter(Negate(is.null), list(p_out, s_out))
    out   <- if (length(parts) == 0L) {
      data.frame(
        Term      = character(0),
        Estimate  = numeric(0),
        Std.Error = numeric(0),
        df        = numeric(0),
        statistic = numeric(0),
        p.value   = numeric(0),
        stringsAsFactors = FALSE
      )
    } else {
      do.call(rbind, parts)
    }

    # ---------------------------------------------------------------------------
    # 2) lm  <-- after gam
    # ---------------------------------------------------------------------------
  } else if (inherits(mod, "lm")) {

    s  <- summary(mod)
    cf <- as.data.frame(s$coefficients)

    out <- data.frame(
      Term      = rownames(cf),
      Estimate  = cf[["Estimate"]],
      Std.Error = cf[["Std. Error"]],
      df        = stats::df.residual(mod),
      statistic = cf[["t value"]],
      p.value   = cf[["Pr(>|t|)"]],
      row.names = NULL,
      stringsAsFactors = FALSE
    )

    # ---------------------------------------------------------------------------
    # 3) lmerMod
    # ---------------------------------------------------------------------------
  } else if (inherits(mod, "lmerMod")) {

    if (isTRUE(pvals) &&
        df_method == "satterthwaite" &&
        requireNamespace("lmerTest", quietly = TRUE)) {

      m_lt <- lmerTest::as_lmerModLmerTest(mod)
      cf   <- as.data.frame(summary(m_lt)$coefficients)

      out <- data.frame(
        Term      = rownames(cf),
        Estimate  = cf[["Estimate"]],
        Std.Error = cf[["Std. Error"]],
        df        = cf[["df"]],
        statistic = cf[["t value"]],
        p.value   = cf[["Pr(>|t|)"]],
        row.names = NULL,
        stringsAsFactors = FALSE
      )

    } else {

      fe <- lme4::fixef(mod)
      se <- sqrt(diag(as.matrix(stats::vcov(mod))))

      out <- data.frame(
        Term      = names(fe),
        Estimate  = unname(fe),
        Std.Error = unname(se),
        df        = NA_real_,
        statistic = unname(fe / se),
        p.value   = NA_real_,
        row.names = NULL,
        stringsAsFactors = FALSE
      )

      if (isTRUE(pvals) &&
          df_method == "satterthwaite" &&
          !requireNamespace("lmerTest", quietly = TRUE)) {
        warning(
          "Install 'lmerTest' to obtain Satterthwaite df and p-values ",
          "for lmerMod.",
          call. = FALSE
        )
      }
    }

    # ---------------------------------------------------------------------------
    # 4) glmerMod
    # ---------------------------------------------------------------------------
  } else if (inherits(mod, "glmerMod")) {

    s  <- summary(mod)
    cf <- as.data.frame(s$coefficients)

    stat_col <- if ("z value" %in% names(cf)) {
      "z value"
    } else if ("t value" %in% names(cf)) {
      "t value"
    } else {
      names(cf)[3L]
    }
    p_col <- if ("Pr(>|z|)" %in% names(cf)) {
      "Pr(>|z|)"
    } else if ("Pr(>|t|)" %in% names(cf)) {
      "Pr(>|t|)"
    } else {
      NA_character_
    }

    out <- data.frame(
      Term      = rownames(cf),
      Estimate  = cf[["Estimate"]],
      Std.Error = cf[["Std. Error"]],
      df        = NA_real_,
      statistic = cf[[stat_col]],
      p.value   = if (!is.na(p_col)) cf[[p_col]] else NA_real_,
      row.names = NULL,
      stringsAsFactors = FALSE
    )

  } else {
    stop(
      "Unknown underlying model class inside `nl_fit`.",
      call. = FALSE
    )
  }

  # ---------------------------------------------------------------------------
  # Drop columns if pvals not requested
  # ---------------------------------------------------------------------------
  if (!isTRUE(pvals)) {
    out$p.value <- NULL
    out$df      <- NULL
  }

  # ---------------------------------------------------------------------------
  # Rounding
  # ---------------------------------------------------------------------------
  out$Estimate  <- maybe_round(out$Estimate)
  out$Std.Error <- maybe_round(out$Std.Error)

  if ("df" %in% names(out)) {
    out$df <- maybe_round(out$df)
  }
  if ("statistic" %in% names(out)) {
    out$statistic <- maybe_round(out$statistic)
  }
  if ("p.value" %in% names(out) && !is.null(digits)) {
    out$p.value <- round(out$p.value, digits + 2L)
  }

  if (requireNamespace("dplyr", quietly = TRUE)) {
    out <- dplyr::as_tibble(out)
  }

  out
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nl_methods.R
\name{print.summary_nl_fit}
\alias{print.summary_nl_fit}
\title{Print method for summary_nl_fit objects}
\usage{
\method{print}{summary_nl_fit}(x, ...)
}
\arguments{
\item{x}{A \code{summary_nl_fit} object.}

\item{...}{Further arguments passed to \code{print}.}
}
\value{
\code{x} invisibly.
}
\description{
Prints a \code{summary_nl_fit} object returned by
\code{\link{summary.nl_fit}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nl_icc.R
\name{nl_icc}
\alias{nl_icc}
\title{Intraclass correlation coefficients for a multilevel nl_fit model}
\usage{
nl_icc(object, include_residual = TRUE)
}
\arguments{
\item{object}{An \code{nl_fit} object returned by \code{\link{nl_fit}}
that was fitted with one or more \code{cluster} variables (i.e., a
multilevel model fitted via \code{lme4::lmer()} or
\code{lme4::glmer()}).}

\item{include_residual}{Logical; if \code{TRUE} (default), includes the
residual variance component \code{ICC_resid} in the output so that all
values sum to 1.}
}
\value{
A named numeric vector of ICCs, one per grouping factor (named
\code{ICC_<groupname>}) plus \code{Residual} for the residual variance
(when \code{include_residual = TRUE}). All values sum to 1.
}
\description{
Extracts variance components from a multilevel \code{nl_fit} object and
computes intraclass correlation coefficients (ICCs) for each grouping
level plus the residual.

The ICC for grouping factor \eqn{g} is defined as:
\deqn{ICC_g = \frac{\sigma^2_g}{\sum_j \sigma^2_j + \sigma^2_\epsilon}}
}
\examples{
\dontrun{
fit <- nl_fit(
  data    = mydata,
  y       = "math_score",
  x       = "SES",
  cluster = c("id", "schid"),
  df      = 4
)
nl_icc(fit)
}

}
\seealso{
\code{\link{nl_fit}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nl_derivatives.R
\name{nl_derivatives}
\alias{nl_derivatives}
\title{Compute first and second derivatives of the fitted spline curve}
\usage{
nl_derivatives(pred_df, x, time = NULL, h = NULL, level = 0.95)
}
\arguments{
\item{pred_df}{A data frame returned by \code{\link{nl_predict}}.}

\item{x}{Character; name of the focal predictor column.}

\item{time}{Optional character; name of the time column. If present,
derivatives are computed separately within each time level.}

\item{h}{Numeric; step size for finite-difference approximation.
Default \code{NULL} uses 0.1 percent of the x range.}

\item{level}{Confidence level for derivative CIs. Default \code{0.95}.}
}
\value{
A data frame with columns \code{x} (the focal predictor),
\code{time} (if applicable), \code{fit} (predicted outcome),
\code{d1} (first derivative), \code{d1_lwr} and \code{d1_upr}
(lower and upper CI for the first derivative), \code{d2} (second
derivative), and \code{d2_lwr} and \code{d2_upr} (CI for the second
derivative).
}
\description{
Uses numerical differentiation on the prediction grid to compute the
first derivative (slope / marginal effect), the second derivative
(curvature), and propagated confidence bands for both, using the
delta method from the \code{se.fit} column of \code{nl_predict()}.
Results are passed to \code{\link{nl_turning_points}} to identify
local turning points and inflection regions.
}
\seealso{
\code{\link{nl_turning_points}}, \code{\link{nl_plot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nl_fit.R
\name{nl_fit}
\alias{nl_fit}
\title{Fit a nonlinear (spline or GAM) single-level or multilevel model}
\usage{
nl_fit(
  data,
  y,
  x,
  time = NULL,
  cluster = NULL,
  nested = FALSE,
  controls = NULL,
  method = c("ns", "bs", "gam"),
  df = 4,
  df_range = 2:8,
  df_criterion = c("AIC", "BIC"),
  k = 5,
  bs_degree = 3,
  random_slope = FALSE,
  family = stats::gaussian(),
  ...
)
}
\arguments{
\item{data}{A data frame (often long format for longitudinal data).}

\item{y}{Outcome variable name (string).}

\item{x}{Focal nonlinear predictor name (string). Must be numeric.}

\item{time}{Optional time variable name (string).}

\item{cluster}{Optional character vector of grouping variable name(s) for
random effects, e.g. \code{NULL}, \code{"id"}, or
\code{c("id", "school")} for two-way clustering.}

\item{nested}{Logical; only used when \code{length(cluster) == 2}. If
\code{TRUE}, uses a nested specification \code{(1 | g1/g2)}; if
\code{FALSE} (default), uses cross-classified
\code{(1 | g1) + (1 | g2)}.}

\item{controls}{Optional character vector of additional covariate names
to include as linear fixed effects.}

\item{method}{Spline basis to use: \code{"ns"} (natural cubic spline,
default), \code{"bs"} (B-spline), or \code{"gam"} (GAM smooth via
\code{mgcv::gam()}). Multilevel fits require \code{"ns"} or \code{"bs"}.}

\item{df}{Degrees of freedom for the spline basis. Supply a single integer
greater than or equal to 1, or the string \code{"auto"} to trigger
automatic selection by information criterion. Default \code{4}.}

\item{df_range}{Integer vector of candidate df values evaluated when
\code{df = "auto"}. Default \code{2:8}.}

\item{df_criterion}{Information criterion used for automatic df selection:
\code{"AIC"} (default) or \code{"BIC"}.}

\item{k}{Basis dimension for \code{mgcv::s()} when \code{method = "gam"}.
Default \code{5}.}

\item{bs_degree}{Polynomial degree for \code{splines::bs()} when
\code{method = "bs"}. Default \code{3} (cubic).}

\item{random_slope}{Logical; if \code{TRUE} and \code{cluster} is supplied,
fits random spline slopes in addition to random intercepts. Currently
implemented only for single-level clustering. Default \code{FALSE}.}

\item{family}{A family object such as \code{stats::gaussian()} (default)
or \code{stats::binomial()}. For multilevel fits, \code{gaussian()} uses
\code{lme4::lmer()} and all other families use \code{lme4::glmer()}.}

\item{...}{Additional arguments passed to the underlying fitting function
(\code{stats::lm()}, \code{mgcv::gam()}, \code{lme4::lmer()}, or
\code{lme4::glmer()}).}
}
\value{
An object of class \code{"nl_fit"} (a named list). It contains
the fitted \code{model} object, the \code{method}, variable names
(\code{y}, \code{x}, \code{time}, \code{cluster}, \code{controls}),
spline settings (\code{df}, \code{df_selected}, \code{df_search},
\code{k}, \code{bs_degree}), flags (\code{nested},
\code{random_slope}), the \code{family}, the model \code{formula},
the \code{call}, and metadata used for prediction (\code{x_info},
\code{levels_info}, \code{control_defaults}).
}
\description{
Fits a nonlinear regression model for an outcome \code{y} with a focal
predictor \code{x}, modeled using natural cubic splines (\code{"ns"}),
B-splines (\code{"bs"}), or GAM smooths (\code{"gam"}).

Version 2 additions: two-way and nested clustering via the
\code{nested} argument; random spline slopes via \code{random_slope};
B-spline basis via \code{method = "bs"}; automatic df selection via
\code{df = "auto"}.
}
\examples{
\dontrun{
# Single-level natural spline with automatic df selection
fit <- nl_fit(data = mydata, y = "score", x = "age", df = "auto")

# Two-way cross-classified clustering
fit2 <- nl_fit(
  data    = mydata,
  y       = "score",
  x       = "age",
  cluster = c("student_id", "school_id"),
  nested  = FALSE
)

# Nested clustering (students within schools)
fit3 <- nl_fit(
  data    = mydata,
  y       = "score",
  x       = "age",
  cluster = c("student_id", "school_id"),
  nested  = TRUE
)

# Random spline slopes
fit4 <- nl_fit(
  data         = mydata,
  y            = "score",
  x            = "age",
  cluster      = "id",
  random_slope = TRUE
)
}

}
\seealso{
\code{\link{nl_predict}}, \code{\link{nl_derivatives}},
\code{\link{nl_compare}}, \code{\link{nl_r2}}, \code{\link{nl_knots}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nl_compare.R
\name{nl_compare}
\alias{nl_compare}
\title{Built-in model comparison workflow}
\usage{
nl_compare(
  object,
  polynomial_degrees = c(2L, 3L),
  digits = 3L,
  return_models = FALSE
)
}
\arguments{
\item{object}{An \code{nl_fit} object (the spline model to compare against).}

\item{polynomial_degrees}{Integer vector of polynomial degrees to include
as additional comparators. Default \code{c(2L, 3L)}. Set to
\code{integer(0)} to skip.}

\item{digits}{Integer; decimal places for the output table. Default \code{3}.}

\item{return_models}{Logical; if \code{TRUE}, also returns the fitted
comparison model objects. Default \code{FALSE}.}
}
\value{
A list (invisibly) of class \code{"nl_compare"} with:
\describe{
\item{\code{table}}{A data frame with the comparison statistics.}
\item{\code{models}}{Named list of fitted models (when
\code{return_models = TRUE}).}
\item{\code{best}}{Name of the model with the lowest AIC.}
}
The table is pretty-printed automatically.
}
\description{
Compares a nonlinear spline model against simpler alternatives—a linear
model and optional polynomial terms—to help researchers justify the
spline approach over simpler specifications.

For each model, the function reports:
\itemize{
\item AIC and BIC
\item Log-likelihood (and likelihood-ratio test versus the linear model
where available)
\item Number of parameters (df)
\item Residual variance / deviance
}
}
\examples{
\dontrun{
fit <- nl_fit(data = mydata, y = "score", x = "age", df = 4)
nl_compare(fit)
nl_compare(fit, polynomial_degrees = c(2, 3, 4))
}

}
\seealso{
\code{\link{nl_fit}}, \code{\link{nl_knots}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nl_predict.R
\name{nl_predict}
\alias{nl_predict}
\title{Generate predictions from an nl_fit model}
\usage{
nl_predict(
  object,
  x_seq = NULL,
  time_levels = NULL,
  controls_fixed = NULL,
  se = TRUE,
  level = 0.95,
  re_form = NA,
  glmer_ci = c("delta", "boot"),
  n_boot = 500L,
  ...
)
}
\arguments{
\item{object}{An \code{nl_fit} object.}

\item{x_seq}{Optional numeric vector of x values. If \code{NULL}, 200
evenly-spaced points between the 1st and 99th percentiles.}

\item{time_levels}{Optional vector of time levels.}

\item{controls_fixed}{Optional named list of fixed control values.}

\item{se}{Logical; include SEs and CIs. Default \code{TRUE}.}

\item{level}{Confidence level. Default \code{0.95}.}

\item{re_form}{For mixed models: \code{NA} (population-level, default) or
\code{NULL} (include random effects).}

\item{glmer_ci}{Method for glmerMod CIs: \code{"delta"} (default) or
\code{"boot"} (parametric bootstrap).}

\item{n_boot}{Number of bootstrap replicates when \code{glmer_ci = "boot"}.
Default \code{500}.}

\item{...}{Reserved for future use.}
}
\value{
A data frame with columns for the focal predictor, time (if any),
controls at fixed values, \code{fit}, \code{se.fit}, \code{lwr},
and \code{upr}.
}
\description{
Creates a prediction data frame over a grid of the focal predictor \code{x}
(and optionally over \code{time}), holding control variables at typical
values. For mixed models, predictions default to population-level curves
(random effects set to zero).

v2 improvements:
\itemize{
\item \strong{CI for glmerMod}: approximate confidence intervals are
computed via the parametric bootstrap or the delta method. Set
\code{glmer_ci = "delta"} (default, fast) or \code{"boot"} (more
accurate, slower).
\item \strong{Cluster-specific predictions}: set \code{re_form = NULL}
to include random effects in the predictions.
}
}
\seealso{
\code{\link{nl_fit}}, \code{\link{nl_plot}},
\code{\link{nl_derivatives}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nl_het.R
\name{nl_het}
\alias{nl_het}
\title{Cluster heterogeneity in nonlinear effects}
\usage{
nl_het(
  object,
  n_clusters_plot = 30L,
  x_seq = NULL,
  level = 0.95,
  plot = TRUE,
  seed = 42L
)
}
\arguments{
\item{object}{An \code{nl_fit} object fitted with a single \code{cluster}
variable (and either \code{random_slope = TRUE} or \code{FALSE}; in the
latter case the model is refit internally).}

\item{n_clusters_plot}{Maximum number of cluster curves to display in the
trajectory plot. Default \code{30L}. Use \code{Inf} for all clusters.}

\item{x_seq}{Optional numeric vector of x values for prediction.}

\item{level}{Confidence level for the population-mean ribbon. Default
\code{0.95}.}

\item{plot}{Logical; print the trajectory plot. Default \code{TRUE}.}

\item{seed}{Integer seed for reproducibility when sub-sampling clusters.
Default \code{42L}.}
}
\value{
A list (invisibly) with:
\describe{
\item{\code{trajectory_df}}{Long data frame of cluster-specific
predicted values.}
\item{\code{slope_variance}}{Named numeric: SD of first-derivative
estimates across clusters at each x value.}
\item{\code{lrt}}{LRT comparing random-slope vs random-intercept model
(\code{NULL} if \code{random_slope = TRUE} was supplied).}
\item{\code{plot}}{A \code{ggplot} object.}
}
}
\description{
Quantifies and visualises how much the nonlinear relationship between
\code{x} and \code{y} varies across clusters, using a model with random
spline slopes.

The function:
\enumerate{
\item Refits the model with random spline slopes (if not already fitted
with \code{random_slope = TRUE}).
\item Extracts cluster-specific predicted curves (BLUPs).
\item Returns a heterogeneity summary: variance of slopes across clusters
at each x value, and a plot of the cluster-specific trajectories.
\item Performs a likelihood-ratio test comparing the random-slope model
against a random-intercept-only model to assess whether heterogeneity
is statistically significant.
}
}
\seealso{
\code{\link{nl_fit}}, \code{\link{nl_derivatives}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nl_summary.R
\name{nl_summary}
\alias{nl_summary}
\title{Tidy coefficient table for an nl_fit model}
\usage{
nl_summary(
  object,
  digits = 3,
  pvals = TRUE,
  df_method = c("satterthwaite", "none")
)
}
\arguments{
\item{object}{An \code{nl_fit} object returned by \code{\link{nl_fit}}.}

\item{digits}{Integer; number of decimal places for rounding. If
\code{NULL}, no rounding is applied. Default \code{3}.}

\item{pvals}{Logical; if \code{TRUE}, attempts to include p-values.
Default \code{TRUE}.}

\item{df_method}{For \code{lmerMod} with \code{pvals = TRUE}:
\code{"satterthwaite"} (requires \pkg{lmerTest}) or \code{"none"}.
Default \code{"satterthwaite"}.}
}
\value{
A data frame (or tibble) with columns: \code{Term}, \code{Estimate},
\code{Std.Error}, \code{df} (if available), \code{statistic}, and
\code{p.value} (if available).
}
\description{
Produces a tidy coefficient table for a model fitted by
\code{\link{nl_fit}}. For linear mixed models (\code{lmerMod}), optional
p-values and denominator degrees of freedom are obtained via
\pkg{lmerTest} (Satterthwaite method). For GLMMs (\code{glmerMod}),
z-tests are reported from \code{summary()}.
}
\seealso{
\code{\link{nl_fit}}, \code{\link{summary.nl_fit}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nl_methods.R
\name{print.nl_fit}
\alias{print.nl_fit}
\title{Print method for nl_fit objects}
\usage{
\method{print}{nl_fit}(x, ...)
}
\arguments{
\item{x}{An object of class \code{nl_fit}.}

\item{...}{Further arguments (currently ignored).}
}
\value{
\code{x} invisibly.
}
\description{
Compact console display for objects returned by \code{\link{nl_fit}}.
Shows key metadata (method, outcome, predictor, time, clustering, family,
and controls) and the fitted model formula.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nl_derivatives.R
\name{nl_turning_points}
\alias{nl_turning_points}
\title{Identify turning points and inflection regions}
\usage{
nl_turning_points(deriv_df, x, time = NULL, tol = 1e-06)
}
\arguments{
\item{deriv_df}{A data frame returned by \code{\link{nl_derivatives}}.}

\item{x}{Character; name of the focal predictor column.}

\item{time}{Optional character; name of the time column. Turning points
are identified separately within each time level when supplied.}

\item{tol}{Numeric tolerance for near-zero detection of d1. Default
\code{1e-6}.}
}
\value{
A named list with three elements.
\code{turning_points} is a data frame with columns \code{x},
\code{time} (if applicable), \code{fit}, and \code{type}
(one of \code{"maximum"}, \code{"minimum"}, or \code{"saddle"}).
\code{inflection_regions} is a data frame with columns
\code{x_start}, \code{x_end}, \code{time}, and \code{direction}
(\code{"concave_to_convex"} or \code{"convex_to_concave"}).
\code{slope_regions} is a data frame with columns \code{x_start},
\code{x_end}, \code{direction}, and \code{time}.
}
\description{
From the derivative table returned by \code{\link{nl_derivatives}}, finds
turning points (values of \code{x} where the first derivative crosses zero),
inflection regions (intervals where the second derivative changes sign),
and slope regions (contiguous stretches where the curve is increasing,
decreasing, or flat).
}
\seealso{
\code{\link{nl_derivatives}}, \code{\link{nl_plot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nl_r2.R
\name{nl_r2}
\alias{nl_r2}
\title{Multilevel R-squared decomposition for nl_fit models}
\usage{
nl_r2(object, digits = 4L)
}
\arguments{
\item{object}{An \code{nl_fit} object returned by \code{\link{nl_fit}}.}

\item{digits}{Integer; decimal places for display. Default \code{4}.}
}
\value{
A list of class \code{"nl_r2"} returned invisibly and
pretty-printed automatically. It contains \code{type} (one of
\code{"OLS"}, \code{"GAM"}, \code{"LMM"}, or \code{"GLMM"}),
\code{r2} (a named numeric vector: \code{R2} and \code{R2_adj} for OLS;
\code{R2_dev} for GAM; \code{R2m} and \code{R2c} for LMM/GLMM), and
\code{variance_partition} (a data frame with columns \code{component},
\code{variance}, and \code{proportion} for multilevel models, or
\code{NULL} for single-level models).
}
\description{
Computes a suite of R-squared statistics for models fitted by
\code{\link{nl_fit}}.
For single-level models the standard R-squared and adjusted R-squared are
returned. For multilevel models (lmerMod / glmerMod) two quantities are
reported: the Nakagawa-Schielzeth marginal R-squared (variance explained by
fixed effects only) and the conditional R-squared (fixed plus all random
effects), together with a level-specific variance partition table analogous
to the r2_mlm / Raudenbush-Bryk approach.
}
\details{
Marginal and conditional R-squared for LMMs follow the Nakagawa and
Schielzeth (2013) formulae extended to multiple random effects by Nakagawa,
Johnson and Schielzeth (2017). The fixed-effects variance
\eqn{\sigma^2_f} is computed as the variance of the linear predictor
from fixed effects only (\eqn{\hat{\mu} = X\hat{\beta}}).

The level-specific variance partition (r2_mlm-style) decomposes the total
modelled variance (\eqn{\sigma^2_f + \sum \sigma^2_j + \sigma^2_\epsilon})
to show how much each source contributes, printed as a breakdown table.
}
\references{
Nakagawa, S., & Schielzeth, H. (2013). A general and simple method for
obtaining R-squared from generalized linear mixed-effects models.
Methods in Ecology and Evolution, 4(2), 133--142.

Nakagawa, S., Johnson, P. C. D., & Schielzeth, H. (2017). The coefficient
of determination R-squared and intra-class correlation coefficient from
generalized linear mixed-effects models revisited and expanded.
Journal of the Royal Society Interface, 14(134), 20170213.

Rights, J. D., & Sterba, S. K. (2019). Quantifying explained variance in
multilevel models: An integrative framework for defining R-squared measures.
Psychological Methods, 24(3), 309--338.
}
\seealso{
\code{\link{nl_fit}}, \code{\link{nl_icc}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nl_knots.R
\name{nl_knots}
\alias{nl_knots}
\title{Automatic knot / degrees-of-freedom selection for spline models}
\usage{
nl_knots(
  data,
  y,
  x,
  time = NULL,
  cluster = NULL,
  nested = FALSE,
  controls = NULL,
  method = c("ns", "bs"),
  df_range = 2:10,
  criterion = c("AIC", "BIC"),
  family = stats::gaussian(),
  plot = TRUE,
  ...
)
}
\arguments{
\item{data}{A data frame.}

\item{y}{Outcome variable name (string).}

\item{x}{Focal predictor name (string).}

\item{time}{Optional time variable name.}

\item{cluster}{Optional character vector of cluster variable names.}

\item{nested}{Logical; nested clustering. Default \code{FALSE}.}

\item{controls}{Optional character vector of control variable names.}

\item{method}{Either \code{"ns"} (default) or \code{"bs"}.}

\item{df_range}{Integer vector of candidate df values. Default \code{2:10}.}

\item{criterion}{Either \code{"AIC"} (default) or \code{"BIC"}.}

\item{family}{A family object. Default \code{stats::gaussian()}.}

\item{plot}{Logical; if \code{TRUE}, prints a diagnostic plot of criterion
vs df. Default \code{TRUE}.}

\item{...}{Additional arguments passed to the underlying fitter.}
}
\value{
A list with:
\describe{
\item{\code{best_df}}{The df value with the lowest criterion value.}
\item{\code{search_table}}{Data frame with columns \code{df} and
\code{criterion}.}
\item{\code{criterion}}{The criterion used (\code{"AIC"} or
\code{"BIC"}).}
\item{\code{plot}}{A \code{ggplot} object (if \code{plot = TRUE}).}
}
}
\description{
Performs a grid search over candidate degrees of freedom (df) for a
natural cubic spline (\code{"ns"}) or B-spline (\code{"bs"}) model and
selects the best value by an information criterion (AIC or BIC). A
diagnostic plot of the criterion against df is returned.

This function is called internally by \code{\link{nl_fit}} when
\code{df = "auto"}, but it can also be called directly for exploration
before fitting the final model.
}
\seealso{
\code{\link{nl_fit}}, \code{\link{nl_compare}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nl_plot.R
\name{nl_plot}
\alias{nl_plot}
\title{Plot predictions and derivatives from nl_fit models}
\usage{
nl_plot(
  pred_df = NULL,
  deriv_df = NULL,
  x,
  time = NULL,
  type = c("trajectory", "slope", "curvature", "combo"),
  show_ci = TRUE,
  ci_level = 0.95,
  show_turning_points = FALSE,
  turning_points = NULL,
  zero_line = TRUE,
  y_lab = NULL,
  x_lab = NULL,
  title = NULL,
  legend_title = NULL
)
}
\arguments{
\item{pred_df}{A data frame from \code{\link{nl_predict}} (for
\code{type = "trajectory"} or \code{"combo"}).}

\item{deriv_df}{A data frame from \code{\link{nl_derivatives}} (for
\code{type = "slope"}, \code{"curvature"}, or \code{"combo"}).}

\item{x}{Character; name of the focal predictor column in \code{pred_df} /
\code{deriv_df}.}

\item{time}{Optional character; name of the time column. Auto-detected if
\code{NULL}.}

\item{type}{Plot type: \code{"trajectory"} (default), \code{"slope"},
\code{"curvature"}, or \code{"combo"}.}

\item{show_ci}{Logical; add confidence ribbons. Default \code{TRUE}.}

\item{ci_level}{Numeric; confidence level when deriving CI from
\code{se.fit}. Default \code{0.95}.}

\item{show_turning_points}{Logical; overlay turning-point markers on the
trajectory plot. Default \code{FALSE}.}

\item{turning_points}{Data frame from \code{\link{nl_turning_points}} (only
used when \code{show_turning_points = TRUE}).}

\item{zero_line}{Logical; for slope / curvature plots, draw a dashed
horizontal line at zero. Default \code{TRUE}.}

\item{y_lab}{Y-axis label.}

\item{x_lab}{X-axis label. Defaults to \code{x}.}

\item{title}{Optional plot title.}

\item{legend_title}{Optional legend title.}
}
\value{
A \code{ggplot} object (or a combined plot list for \code{"combo"}).
}
\description{
Visualises results from \code{\link{nl_predict}} or
\code{\link{nl_derivatives}}.  The \code{type} argument selects the plot:
\itemize{
\item \code{"trajectory"}: predicted outcome vs \code{x} with optional
confidence ribbon (v1 behaviour, now default).
\item \code{"slope"}: first derivative (marginal effect) vs \code{x}.
\item \code{"curvature"}: second derivative vs \code{x}.
\item \code{"combo"}: trajectory + slope panels side by side.
}
}
\seealso{
\code{\link{nl_predict}}, \code{\link{nl_derivatives}},
\code{\link{nl_turning_points}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nl_methods.R
\name{summary.nl_fit}
\alias{summary.nl_fit}
\title{Summary method for nl_fit objects}
\usage{
\method{summary}{nl_fit}(
  object,
  digits = 3,
  pvals = TRUE,
  df_method = c("satterthwaite", "none"),
  ...
)
}
\arguments{
\item{object}{An \code{nl_fit} object returned by \code{\link{nl_fit}}.}

\item{digits}{Number of decimal places for rounding. Default \code{3}.}

\item{pvals}{Logical; if \code{TRUE}, attempts to include p-values.
Default \code{TRUE}.}

\item{df_method}{For \code{lmerMod}: \code{"satterthwaite"} (requires
\pkg{lmerTest}) or \code{"none"}. Default \code{"satterthwaite"}.}

\item{...}{Further arguments passed to \code{\link{nl_summary}}.}
}
\value{
An object of class \code{summary_nl_fit} containing \code{call},
\code{formula}, \code{method}, and \code{table}.
}
\description{
Produces a tidy coefficient table via \code{\link{nl_summary}} and wraps
it in a \code{summary_nl_fit} object for pretty printing.
}
